#
# Basic Single Particle Model (SPM)
#
import pybamm

from .base_lithium_ion_model import BaseModel


class BasicSPM(BaseModel):
    """Single Particle Model (SPM) model of a lithium-ion battery, from
    :footcite:t:`Marquis2019`.

    This class differs from the :class:`pybamm.lithium_ion.SPM` model class in that it
    shows the whole model in a single class. This comes at the cost of flexibility in
    combining different physical effects, and in general the main SPM class should be
    used instead.

    Parameters
    ----------
    name : str, optional
        The name of the model.
    """

    def __init__(self, name="Single Particle Model"):
        super().__init__({}, name)
        pybamm.citations.register("Marquis2019")
        # `param` is a class containing all the relevant parameters and functions for
        # this model. These are purely symbolic at this stage, and will be set by the
        # `ParameterValues` class when the model is processed.

        ######################
        # Variables
        ######################
        # Variables that depend on time only are created without a domain
        Q = pybamm.Variable("Discharge capacity [A.h]")
        # Variables that vary spatially are created with a domain
        c_s_n = pybamm.Variable(
            "X-averaged negative particle concentration [mol.m-3]",
            domain="negative particle",
        )
        c_s_p = pybamm.Variable(
            "X-averaged positive particle concentration [mol.m-3]",
            domain="positive particle",
        )

        # Constant temperature
        T = self.param.T_init

        ######################
        # Other set-up
        ######################

        # Current density
        i_cell = self.param.current_density_with_time
        a_n = 3 * self.param.n.prim.epsilon_s_av / self.param.n.prim.R_typ
        a_p = 3 * self.param.p.prim.epsilon_s_av / self.param.p.prim.R_typ
        j_n = i_cell / (self.param.n.L * a_n)
        j_p = -i_cell / (self.param.p.L * a_p)

        ######################
        # State of Charge
        ######################
        I = self.param.current_with_time
        # The `rhs` dictionary contains differential equations, with the key being the
        # variable in the d/dt
        self.rhs[Q] = I / 3600
        # Initial conditions must be provided for the ODEs
        self.initial_conditions[Q] = pybamm.Scalar(0)

        ######################
        # Particles
        ######################

        # The div and grad operators will be converted to the appropriate matrix
        # multiplication at the discretisation stage
        N_s_n = -self.param.n.prim.D(c_s_n, T) * pybamm.grad(c_s_n)
        N_s_p = -self.param.p.prim.D(c_s_p, T) * pybamm.grad(c_s_p)
        self.rhs[c_s_n] = -pybamm.div(N_s_n)
        self.rhs[c_s_p] = -pybamm.div(N_s_p)
        # Surf takes the surface value of a variable, i.e. its boundary value on the
        # right side. This is also accessible via `boundary_value(x, "right")`, with
        # "left" providing the boundary value of the left side
        c_s_surf_n = pybamm.surf(c_s_n)
        c_s_surf_p = pybamm.surf(c_s_p)
        # Boundary conditions must be provided for equations with spatial derivatives
        self.boundary_conditions[c_s_n] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -j_n / (self.param.F * pybamm.surf(self.param.n.prim.D(c_s_n, T))),
                "Neumann",
            ),
        }
        self.boundary_conditions[c_s_p] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -j_p / (self.param.F * pybamm.surf(self.param.p.prim.D(c_s_p, T))),
                "Neumann",
            ),
        }
        # c_n_init and c_p_init are functions of r and x, but for the SPM we
        # take the x-averaged value since there is no x-dependence in the particles
        self.initial_conditions[c_s_n] = pybamm.x_average(self.param.n.prim.c_init)
        self.initial_conditions[c_s_p] = pybamm.x_average(self.param.p.prim.c_init)
        # Events specify points at which a solution should terminate
        sto_surf_n = c_s_surf_n / self.param.n.prim.c_max
        sto_surf_p = c_s_surf_p / self.param.p.prim.c_max
        self.events += [
            pybamm.Event(
                "Minimum negative particle surface stoichiometry",
                pybamm.min(sto_surf_n) - 0.01,
            ),
            pybamm.Event(
                "Maximum negative particle surface stoichiometry",
                (1 - 0.01) - pybamm.max(sto_surf_n),
            ),
            pybamm.Event(
                "Minimum positive particle surface stoichiometry",
                pybamm.min(sto_surf_p) - 0.01,
            ),
            pybamm.Event(
                "Maximum positive particle surface stoichiometry",
                (1 - 0.01) - pybamm.max(sto_surf_p),
            ),
        ]

        # Note that the SPM does not have any algebraic equations, so the `algebraic`
        # dictionary remains empty

        ######################
        # (Some) variables
        ######################
        # Interfacial reactions
        RT_F = self.param.R * T / self.param.F
        j0_n = self.param.n.prim.j0(self.param.c_e_init_av, c_s_surf_n, T)
        j0_p = self.param.p.prim.j0(self.param.c_e_init_av, c_s_surf_p, T)
        eta_n = (2 / self.param.n.prim.ne) * RT_F * pybamm.arcsinh(j_n / (2 * j0_n))
        eta_p = (2 / self.param.p.prim.ne) * RT_F * pybamm.arcsinh(j_p / (2 * j0_p))
        phi_s_n = 0
        phi_e = -eta_n - self.param.n.prim.U(sto_surf_n, T)
        phi_s_p = eta_p + phi_e + self.param.p.prim.U(sto_surf_p, T)
        V = phi_s_p
        num_cells = pybamm.Parameter(
            "Number of cells connected in series to make a battery"
        )

        whole_cell = ["negative electrode", "separator", "positive electrode"]
        # The `variables` dictionary contains all variables that might be useful for
        # visualising the solution of the model
        # Primary broadcasts are used to broadcast scalar quantities across a domain
        # into a vector of the right shape, for multiplying with other vectors
        self.variables = {
            "Time [s]": pybamm.t,
            "Discharge capacity [A.h]": Q,
            "X-averaged negative particle concentration [mol.m-3]": c_s_n,
            "Negative particle surface "
            "concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                c_s_surf_n, "negative electrode"
            ),
            "Electrolyte concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                self.param.c_e_init_av, whole_cell
            ),
            "X-averaged positive particle concentration [mol.m-3]": c_s_p,
            "Positive particle surface "
            "concentration [mol.m-3]": pybamm.PrimaryBroadcast(
                c_s_surf_p, "positive electrode"
            ),
            "Current [A]": I,
            "Current variable [A]": I,  # for compatibility with pybamm.Experiment
            "Negative electrode potential [V]": pybamm.PrimaryBroadcast(
                phi_s_n, "negative electrode"
            ),
            "Electrolyte potential [V]": pybamm.PrimaryBroadcast(phi_e, whole_cell),
            "Positive electrode potential [V]": pybamm.PrimaryBroadcast(
                phi_s_p, "positive electrode"
            ),
            "Voltage [V]": V,
            "Battery voltage [V]": V * num_cells,
        }
        # Events specify points at which a solution should terminate
        self.events += [
            pybamm.Event("Minimum voltage [V]", V - self.param.voltage_low_cut),
            pybamm.Event("Maximum voltage [V]", self.param.voltage_high_cut - V),
        ]



import numpy as np

import pybamm


def set_initial_state(
    initial_value,
    parameter_values,
    direction=None,
    param=None,
    inplace=True,
    options=None,
    inputs=None,
    tol=1e-6,
):
    """
    Set the value of the initial state of charge.

    Parameters
    ----------
    initial_value : float
        Target initial value.
        If float, interpreted as SOC, must be between 0 and 1.
        If string e.g. "4 V", interpreted as voltage, must be between V_min and V_max.
    parameter_values : :class:`pybamm.ParameterValues`
        Parameters and their corresponding values.
    param : :class:`pybamm.LithiumIonParameters`, optional
        The symbolic parameter set to use for the simulation.
        If not provided, the default parameter set will be used.
    inplace: bool, optional
        If True, replace the parameters values in place. Otherwise, return a new set of
        parameter values. Default is True.
    options : dict-like, optional
        A dictionary of options to be passed to the model, see
        :class:`pybamm.BatteryModelOptions`.
    inputs : dict, optional
        A dictionary of input parameters to pass to the model when solving.
    tol : float, optional
        The tolerance for the solver used to compute the initial stoichiometries.
        A lower value results in higher precision but may increase computation time.
        Default is 1e-6.
    """
    parameter_values = parameter_values if inplace else parameter_values.copy()
    param = param or pybamm.EcmParameters()

    if isinstance(initial_value, str) and initial_value.endswith("V"):
        V_init = float(initial_value[:-1])
        V_min = parameter_values.evaluate(param.voltage_low_cut, inputs=inputs)
        V_max = parameter_values.evaluate(param.voltage_high_cut, inputs=inputs)

        if not V_min <= V_init <= V_max:
            raise ValueError(
                f"Initial voltage {V_init}V is outside the voltage limits "
                f"({V_min}, {V_max})"
            )

        # Solve simple model for initial soc based on target voltage
        soc_model = pybamm.BaseModel()
        soc = pybamm.Variable("soc")
        soc_model.algebraic[soc] = param.ocv(soc) - V_init

        # initial guess for soc linearly interpolates between 0 and 1
        # based on V linearly interpolating between V_max and V_min
        soc_model.initial_conditions[soc] = (V_init - V_min) / (V_max - V_min)
        soc_model.variables["soc"] = soc
        parameter_values.process_model(soc_model)
        initial_soc = (
            pybamm.AlgebraicSolver(tol=tol)
            .solve(soc_model, [0], inputs=inputs)["soc"]
            .data[0]
        )

        # Ensure that the result lies between 0 and 1
        parameter_values["Initial SoC"] = np.minimum(np.maximum(initial_soc, 0.0), 1.0)

    elif isinstance(initial_value, int | float):
        if not 0 <= initial_value <= 1:
            raise ValueError("Initial SOC should be between 0 and 1")
        parameter_values["Initial SoC"] = initial_value

    else:
        raise ValueError(
            "Initial value must be a float between 0 and 1, or a string ending in 'V'"
        )

    return parameter_values



#
# Spatial method for zero dimensional meshes
#
import numpy as np

import pybamm


class ZeroDimensionalSpatialMethod(pybamm.SpatialMethod):
    """
    A discretisation class for the zero dimensional mesh

    Parameters
    ----------
    mesh : :class: `pybamm.Mesh`
        Contains all the submeshes for discretisation
    """

    def __init__(self, options=None):
        super().__init__(options)

    def build(self, mesh):
        self._mesh = mesh

    def boundary_value_or_flux(self, symbol, discretised_child, bcs=None):
        """
        In 0D, the boundary value is the identity operator.
        See :meth:`SpatialMethod.boundary_value_or_flux`
        """
        return discretised_child

    def mass_matrix(self, symbol, boundary_conditions):
        """
        Calculates the mass matrix for a spatial method. Since the spatial method is
        zero dimensional, this is simply the number 1.
        """
        return pybamm.Matrix(np.ones((1, 1)))

    def indefinite_integral(self, child, discretised_child, direction):
        """
        Calculates the zero-dimensional indefinite integral.
        If 'direction' is forward, this is the identity operator.
        If 'direction' is backward, this is the negation operator.
        """
        if direction == "forward":
            return discretised_child
        elif direction == "backward":
            return -discretised_child

    def integral(
        self, child, discretised_child, integration_dimension, integration_variable
    ):
        """
        Calculates the zero-dimensional integral, i.e. the identity operator
        """
        return discretised_child



__all__ = ['algebraic_solver', 'base_solver', 'c_solvers',
           'casadi_algebraic_solver', 'casadi_solver', 'dummy_solver',
           'idaklu_jax', 'idaklu_solver', 'jax_bdf_solver', 'jax_solver',
           'lrudict', 'processed_variable', 'processed_variable_computed',
           'scipy_solver', 'solution', 'processed_variable_time_integral']



#
# Interpolating class
#
from __future__ import annotations

import numbers
from collections.abc import Sequence
from typing import Any

import numpy as np
import numpy.typing as npt
from scipy import interpolate

import pybamm


class Interpolant(pybamm.Function):
    """
    Interpolate data in 1D, 2D, or 3D. Interpolation in 3D requires the input data to be
    on a regular grid (as per scipy.interpolate.RegularGridInterpolator).

    Parameters
    ----------
    x : iterable of :class:`numpy.ndarray`
        The data point coordinates. If 1-D, then this is an array(s) of real values. If,
        2D or 3D interpolation, then this is to ba a tuple of 1D arrays (one for each
        dimension) which together define the coordinates of the points.
    y : :class:`numpy.ndarray`
        The values of the function to interpolate at the data points. In 2D and 3D, this
        should be a matrix of two and three dimensions respectively.
    children : iterable of :class:`pybamm.Symbol`
        Node(s) to use when evaluating the interpolant. Each child corresponds to an
        entry of x
    name : str, optional
        Name of the interpolant. Default is None, in which case the name "interpolating
        function" is given.
    interpolator : str, optional
        Which interpolator to use. Can be "linear", "cubic", or "pchip". Default is
        "linear". For 3D interpolation, only "linear" an "cubic" are currently
        supported.
    extrapolate : bool, optional
        Whether to extrapolate for points that are outside of the parametrisation
        range, or return NaN (following default behaviour from scipy). Default is True.
        Generally, it is best to set this to be False for 3D interpolation due to
        the higher potential for errors in extrapolation.
    """

    def __init__(
        self,
        x: npt.NDArray[np.float64] | Sequence[npt.NDArray[np.float64]],
        y: npt.NDArray[Any],
        children: Sequence[pybamm.Symbol] | pybamm.Time,
        name: str | None = None,
        interpolator: str | None = "linear",
        extrapolate: bool = True,
        entries_string: str | None = None,
        _num_derivatives: int = 0,
    ):
        # Check interpolator is valid
        if interpolator not in ["linear", "cubic", "pchip"]:
            raise ValueError(f"interpolator '{interpolator}' not recognised")

        # Perform some checks on the data
        if isinstance(x, tuple | list) and len(x) == 2:
            x1, x2 = x
            if y.ndim != 2:
                raise ValueError("y should be two-dimensional if len(x)=2")
            if x1.shape[0] != y.shape[0]:
                raise ValueError(
                    "len(x1) should equal y=shape[1], "
                    f"but x1.shape={x1.shape} and y.shape={y.shape}"
                )
            if x2 is not None and x2.shape[0] != y.shape[1]:
                raise ValueError(
                    "len(x2) should equal y=shape[0], "
                    f"but x2.shape={x2.shape} and y.shape={y.shape}"
                )
        elif isinstance(x, tuple | list) and len(x) == 3:
            x1, x2, x3 = x
            if y.ndim != 3:
                raise ValueError("y should be three-dimensional if len(x)=3")

            if x1.shape[0] != y.shape[0]:
                raise ValueError(
                    "len(x1) should equal y=shape[0], "
                    f"but x1.shape={x1.shape} and y.shape={y.shape}"
                )
            if x2 is not None and x2.shape[0] != y.shape[1]:
                raise ValueError(
                    "len(x2) should equal y=shape[1], "
                    f"but x2.shape={x2.shape} and y.shape={y.shape}"
                )
            if x3 is not None and x3.shape[0] != y.shape[2]:
                raise ValueError(
                    "len(x3) should equal y=shape[2], "
                    f"but x3.shape={x3.shape} and y.shape={y.shape}"
                )
        else:
            if isinstance(x, tuple | list):
                x1 = x[0]
            else:
                x1 = x
                x: list[npt.NDArray[np.float64]] = [x]  # type: ignore[no-redef]
            x2 = None
            if x1.shape[0] != y.shape[0]:
                raise ValueError(
                    "len(x1) should equal y=shape[0], "
                    f"but x1.shape={x1.shape} and y.shape={y.shape}"
                )
            # if 1d then x should be monotonically increasing
            if np.any(x1[:-1] > x1[1:]):
                raise ValueError("x should be monotonically increasing")
        # children should be a list not a symbol or a number
        if isinstance(children, pybamm.Symbol | numbers.Number):
            children = [children]
        # Either a single x is provided and there is one child
        # or x is a 2-tuple and there are two children
        if len(x) != len(children):
            raise ValueError("len(x) should equal len(children)")
        # if there is only one x, y can be 2-dimensional but the child must have
        # length 1
        if len(x) == 1 and y.ndim == 2 and children[0].size != 1:
            raise ValueError(
                "child should have size 1 if y is two-dimensional and len(x)==1"
            )

        # Create interpolating function
        if len(x) == 1:
            self.dimension = 1
            if interpolator == "linear":
                if extrapolate is False:
                    fill_value_1: float | str = np.nan
                elif extrapolate is True:
                    fill_value_1 = "extrapolate"
                interpolating_function = interpolate.interp1d(
                    x1,
                    y,
                    bounds_error=False,
                    fill_value=fill_value_1,
                    axis=0,
                )
            elif interpolator == "cubic":
                interpolating_function = interpolate.CubicSpline(
                    x1, y, extrapolate=extrapolate
                )
            elif interpolator == "pchip":
                interpolating_function = interpolate.PchipInterpolator(
                    x1, y, extrapolate=extrapolate
                )
        elif len(x) == 2:
            self.dimension = 2
            if interpolator == "pchip":
                raise ValueError(
                    "interpolator should be 'linear' or 'cubic' if x is two-dimensional"
                )
            else:
                if extrapolate:
                    fill_value = None
                else:
                    fill_value = np.nan
                interpolating_function = interpolate.RegularGridInterpolator(
                    (x1, x2),
                    y,
                    method=interpolator,
                    bounds_error=False,
                    fill_value=fill_value,
                )

        elif len(x) == 3:
            self.dimension = 3

            if extrapolate:
                fill_value = None
            else:
                fill_value = np.nan

            possible_interpolators = ["linear", "cubic"]
            if interpolator not in possible_interpolators:
                raise ValueError(
                    """interpolator should be 'linear' or 'cubic'
                    for 3D interpolation"""
                )
            else:
                interpolating_function = interpolate.RegularGridInterpolator(
                    (x1, x2, x3),
                    y,
                    method=interpolator,
                    bounds_error=False,
                    fill_value=fill_value,
                )
        else:
            raise ValueError(f"Invalid dimension of x: {len(x)}")

        # Set name
        if name is None:
            name = "interpolating_function"
        self.x = x
        self.y = y
        self.entries_string = entries_string

        # Differentiate the interpolating function if necessary
        self._num_derivatives = _num_derivatives
        for _ in range(_num_derivatives):
            interpolating_function = interpolating_function.derivative()

        super().__init__(interpolating_function, *children, name=name)

        # Store information as attributes
        self.interpolator = interpolator
        self.extrapolate = extrapolate

    @classmethod
    def _from_json(cls, snippet: dict):
        """Create an Interpolant object from JSON data"""

        x1 = []

        if len(snippet["x"]) == 1:
            x1 = [np.array(x) for x in snippet["x"]]

        return cls(
            x1 if x1 else tuple(np.array(x) for x in snippet["x"]),
            np.array(snippet["y"]),
            snippet["children"],
            name=snippet["name"],
            interpolator=snippet["interpolator"],
            extrapolate=snippet["extrapolate"],
            _num_derivatives=snippet["_num_derivatives"],
        )

    @property
    def entries_string(self):
        return self._entries_string

    @entries_string.setter
    def entries_string(self, value):
        # We must include the entries in the hash, since different arrays can be
        # indistinguishable by class, name and domain alone
        # Slightly different syntax for sparse and non-sparse matrices
        if value is not None:
            self._entries_string = value
        else:
            self._entries_string = ""
            for i, x in enumerate(self.x):
                self._entries_string += "x" + str(i) + "_" + str(x.tobytes())
            self._entries_string += "y_" + str(self.y.tobytes())

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`."""
        self._id = hash(
            (
                self.__class__,
                self.name,
                self.entries_string,
                *tuple([child.id for child in self.children]),
                *tuple(self.domain),
                self._num_derivatives,
            )
        )

    def create_copy(self, new_children=None, perform_simplifications=True):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        children = self._children_for_copying(new_children)

        return pybamm.Interpolant(
            self.x,
            self.y,
            children,
            name=self.name,
            interpolator=self.interpolator,
            extrapolate=self.extrapolate,
            entries_string=self.entries_string,
            _num_derivatives=self._num_derivatives,
        )

    def _function_evaluate(self, evaluated_children):
        children_eval_flat = []
        for child in evaluated_children:
            if isinstance(child, np.ndarray):
                children_eval_flat.append(child.flatten())
            else:
                children_eval_flat.append(child)
        if self.dimension == 1:
            return self.function(*children_eval_flat).flatten()[:, np.newaxis]
        elif self.dimension in [2, 3]:
            # If the children are scalars, we need to add a dimension
            shapes = []
            for child in evaluated_children:
                if isinstance(child, float | int):
                    shapes.append(())
                else:
                    shapes.append(child.shape)
            shapes = set(shapes)
            shapes.discard(())

            if len(shapes) > 1:
                raise ValueError(
                    "All children must have the same shape for 3D interpolation"
                )

            if len(shapes) == 0:
                shape = (1,)
            else:
                shape = shapes.pop()
            new_evaluated_children = []
            for child in evaluated_children:
                if hasattr(child, "shape") and child.shape == shape:
                    new_evaluated_children.append(child.flatten())
                else:
                    new_evaluated_children.append(np.reshape(child, shape).flatten())

            # return nans if there are any within the children
            nans = np.isnan(new_evaluated_children)
            if np.any(nans):
                nan_children = []
                for child, interp_range in zip(
                    new_evaluated_children, self.function.grid, strict=True
                ):
                    nan_children.append(np.ones_like(child) * interp_range.mean())
                nan_eval = self.function(np.transpose(nan_children))
                return np.reshape(nan_eval, shape)
            else:
                res = self.function(np.transpose(new_evaluated_children))
                return np.reshape(res, shape)

        else:  # pragma: no cover
            raise ValueError(f"Invalid dimension: {self.dimension}")

    def _function_diff(self, children: Sequence[pybamm.Symbol], idx: float):
        """
        Derivative with respect to child number 'idx'.
        See :meth:`pybamm.Symbol._diff()`.
        """
        if len(children) > 1:
            raise NotImplementedError(
                "differentiation not implemented for functions with more than one child"
            )
        else:
            # keep using "derivative" as derivative
            return Interpolant(
                self.x,
                self.y,
                children,
                name=self.name,
                interpolator=self.interpolator,
                extrapolate=self.extrapolate,
                _num_derivatives=self._num_derivatives + 1,
            )

    def to_json(self):
        """
        Method to serialise an Interpolant object into JSON.
        """

        json_dict = {
            "name": self.name,
            "id": self.id,
            "x": [x_item.tolist() for x_item in self.x],
            "y": self.y.tolist(),
            "interpolator": self.interpolator,
            "extrapolate": self.extrapolate,
            "_num_derivatives": self._num_derivatives,
        }

        return json_dict



from __future__ import annotations

import copy
import numbers
import warnings
from collections import OrderedDict

import casadi
import numpy as np
import scipy

import pybamm
from pybamm.expression_tree.operations.serialise import Serialise


class BaseModel:
    """
    Base model class for other models to extend.

    Attributes
    ----------
    name: str
        A string representing the name of the model.
    submodels: dict
        A dictionary of submodels that the model is composed of.
    use_jacobian: bool
        Whether to use the Jacobian when solving the model (default is True).
    convert_to_format: str
        Specifies the format to convert the expression trees representing the RHS,
        algebraic equations, Jacobian, and events.
        Options are:

            - None: retain PyBaMM expression tree structure.
            - "python": convert to Python code for evaluating `evaluate(t, y)` on expressions.
            - "casadi": convert to CasADi expression tree for Jacobian calculation.
            - "jax": convert to JAX expression tree.

        Default is "casadi".
    is_discretised: bool
        Indicates whether the model has been discretised (default is False).
    y_slices: None or list
        Slices of the concatenated state vector after discretisation, used to track
        different submodels in the full concatenated solution vector.
    """

    def __init__(self, name="Unnamed model"):
        self.name = name
        self._options = {}
        self._built = False
        self._built_fundamental = False

        # Initialise empty model
        self.submodels = {}
        self._rhs = {}
        self._algebraic = {}
        self._initial_conditions = {}
        self._boundary_conditions = {}
        self._variables_by_submodel = {}
        self._variables = pybamm.FuzzyDict({})
        self._coupled_variables = {}
        self._summary_variables = []
        self._events = []
        self._concatenated_rhs = None
        self._concatenated_algebraic = None
        self._concatenated_initial_conditions = None
        self._mass_matrix = None
        self._mass_matrix_inv = None
        self._jacobian = None
        self._jacobian_algebraic = None
        self._parameters = None
        self._input_parameters = None
        self._parameter_info = None
        self._is_standard_form_dae = None
        self._variables_casadi = {}
        self._geometry = pybamm.Geometry({})

        # Default behaviour is to use the jacobian
        self.use_jacobian = True
        self.convert_to_format = "casadi"

        # Model is not initially discretised
        self.is_discretised = False
        self.y_slices = None
        self.len_rhs_and_alg = None

        # Non-lithium ion models shouldn't calculate eSOH parameters
        self._calc_esoh = False

        # Root solver
        self._algebraic_root_solver = None

    @classmethod
    def deserialise(cls, properties: dict):
        """
        Create a model instance from a serialised object.
        """

        # append the model name with _saved to differentiate
        instance = cls(name=properties["name"] + "_saved")

        instance.options = properties["options"]

        return cls.generic_deserialise(instance, properties)

    @classmethod
    def generic_deserialise(cls, instance, properties):
        # Initialise model with stored variables that have already been discretised
        instance._concatenated_rhs = properties["concatenated_rhs"]
        instance._concatenated_algebraic = properties["concatenated_algebraic"]
        instance._concatenated_initial_conditions = properties[
            "concatenated_initial_conditions"
        ]
        instance.len_rhs = instance.concatenated_rhs.size
        instance.len_alg = instance.concatenated_algebraic.size
        instance.len_rhs_and_alg = instance.len_rhs + instance.len_alg
        instance.bounds = properties["bounds"]
        instance.events = properties["events"]
        instance.mass_matrix = properties["mass_matrix"]
        instance.mass_matrix_inv = properties["mass_matrix_inv"]
        # add optional properties not required for model to solve
        if properties["variables"]:
            instance._variables = pybamm.FuzzyDict(properties["variables"])

            # assign meshes to each variable
            for var in instance._variables.values():
                if var.domain != []:
                    var.mesh = properties["mesh"][var.domain]

                if var.domains["secondary"] != []:
                    var.secondary_mesh = properties["mesh"][var.domains["secondary"]]
                else:
                    var.secondary_mesh = None

                if var.domains["tertiary"] != []:
                    var.tertiary_mesh = properties["mesh"][var.domains["tertiary"]]
                else:
                    var.tertiary_mesh = None

            if properties["geometry"]:
                instance._geometry = pybamm.Geometry(properties["geometry"])
        else:
            # Delete the default variables which have not been discretised
            instance._variables = pybamm.FuzzyDict({})
        # Model has already been discretised
        instance.is_discretised = True
        return instance

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def rhs(self):
        """Returns a dictionary mapping expressions (variables) to expressions that represent
        the right-hand side (RHS) of the model's differential equations."""
        return self._rhs

    @rhs.setter
    def rhs(self, rhs):
        self._rhs = EquationDict("rhs", rhs)

    @property
    def algebraic(self):
        """Returns a dictionary mapping expressions (variables) to expressions that represent
        the algebraic equations of the model."""
        return self._algebraic

    @algebraic.setter
    def algebraic(self, algebraic):
        self._algebraic = EquationDict("algebraic", algebraic)

    @property
    def initial_conditions(self):
        """Returns a dictionary mapping expressions (variables) to expressions that represent
        the initial conditions for the state variables."""
        return self._initial_conditions

    @initial_conditions.setter
    def initial_conditions(self, initial_conditions):
        self._initial_conditions = EquationDict(
            "initial_conditions", initial_conditions
        )

    @property
    def boundary_conditions(self):
        """Returns a dictionary mapping expressions (variables) to expressions representing
        the boundary conditions of the model."""
        return self._boundary_conditions

    @boundary_conditions.setter
    def boundary_conditions(self, boundary_conditions):
        self._boundary_conditions = BoundaryConditionsDict(boundary_conditions)

    @property
    def coupled_variables(self):
        """Returns a dictionary mapping strings to expressions representing variables needed by the model but whose equations were set by other models."""
        return self._coupled_variables

    @coupled_variables.setter
    def coupled_variables(self, coupled_variables):
        for name, var in coupled_variables.items():
            if (
                isinstance(var, pybamm.CoupledVariable)
                and var.name != name
                # Exception if the variable is also there under its own name
                and not (
                    var.name in coupled_variables and coupled_variables[var.name] == var
                )
            ):
                raise ValueError(
                    f"Coupled variable with name '{var.name}' is in coupled variables dictionary with "
                    f"name '{name}'. Names must match."
                )
        self._coupled_variables = coupled_variables

    def list_coupled_variables(self):
        return list(self._coupled_variables.keys())

    @property
    def variables(self):
        """Returns a dictionary mapping strings to expressions representing the model's useful variables."""
        return self._variables

    @variables.setter
    def variables(self, variables):
        for name, var in variables.items():
            if (
                isinstance(var, pybamm.Variable)
                and var.name != name
                # Exception if the variable is also there under its own name
                and not (var.name in variables and variables[var.name] == var)
            ):
                raise ValueError(
                    f"Variable with name '{var.name}' is in variables dictionary with "
                    f"name '{name}'. Names must match."
                )
        self._variables = pybamm.FuzzyDict(variables)

    def variable_names(self):
        return list(self._variables.keys())

    @property
    def variables_and_events(self):
        """Returns a dictionary containing both models variables and events."""
        try:
            return self._variables_and_events
        except AttributeError:
            self._variables_and_events = self.variables.copy()
            self._variables_and_events.update(
                {f"Event: {event.name}": event.expression for event in self.events}
            )
            return self._variables_and_events

    @property
    def events(self):
        """Returns a dictionary mapping expressions (variables) to expressions that represent
        the initial conditions for the state variables."""
        return self._events

    @events.setter
    def events(self, events):
        self._events = events

    @property
    def concatenated_rhs(self):
        """Returns the concatenated right-hand side (RHS) expressions for the model after discretisation."""
        return self._concatenated_rhs

    @concatenated_rhs.setter
    def concatenated_rhs(self, concatenated_rhs):
        self._concatenated_rhs = concatenated_rhs

    @property
    def concatenated_algebraic(self):
        """Returns the concatenated algebraic equations for the model after discretisation."""
        return self._concatenated_algebraic

    @concatenated_algebraic.setter
    def concatenated_algebraic(self, concatenated_algebraic):
        self._concatenated_algebraic = concatenated_algebraic

    @property
    def concatenated_initial_conditions(self):
        """Returns the initial conditions for all variables after discretization, providing the
        initial values for the state variables."""
        return self._concatenated_initial_conditions

    @concatenated_initial_conditions.setter
    def concatenated_initial_conditions(self, concatenated_initial_conditions):
        self._concatenated_initial_conditions = concatenated_initial_conditions

    @property
    def built(self):
        "Returns a boolean for the model built status."
        return self._built

    @property
    def mass_matrix(self):
        """Returns the mass matrix for the system of differential equations after discretisation."""
        return self._mass_matrix

    @mass_matrix.setter
    def mass_matrix(self, mass_matrix):
        self._mass_matrix = mass_matrix

    @property
    def mass_matrix_inv(self):
        """Returns the inverse of the mass matrix for the differential equations after discretisation."""
        return self._mass_matrix_inv

    @mass_matrix_inv.setter
    def mass_matrix_inv(self, mass_matrix_inv):
        self._mass_matrix_inv = mass_matrix_inv

    @property
    def jacobian(self):
        """Returns the Jacobian matrix for the model, computed automatically if `use_jacobian` is True."""
        return self._jacobian

    @jacobian.setter
    def jacobian(self, jacobian):
        self._jacobian = jacobian

    @property
    def jacobian_rhs(self):
        """Returns the Jacobian matrix for the right-hand side (RHS) part of the model, computed
        if `use_jacobian` is True."""
        return self._jacobian_rhs

    @jacobian_rhs.setter
    def jacobian_rhs(self, jacobian_rhs):
        self._jacobian_rhs = jacobian_rhs

    @property
    def jacobian_algebraic(self):
        """Returns the Jacobian matrix for the algebraic part of the model, computed automatically
        during solver setup if `use_jacobian` is True."""
        return self._jacobian_algebraic

    @jacobian_algebraic.setter
    def jacobian_algebraic(self, jacobian_algebraic):
        self._jacobian_algebraic = jacobian_algebraic

    @property
    def param(self):
        """Returns a dictionary to store parameter values for the model."""
        return self._param

    @param.setter
    def param(self, values):
        self._param = values

    @property
    def options(self):
        """Returns the model options dictionary that allows customization of the model's behavior."""
        return self._options

    @options.setter
    def options(self, options):
        self._options = options

    @property
    def timescale(self):
        raise NotImplementedError(
            "timescale has been removed since models are now dimensional"
        )

    @timescale.setter
    def timescale(self, value):
        raise NotImplementedError(
            "timescale has been removed since models are now dimensional"
        )

    @property
    def length_scales(self):
        raise NotImplementedError(
            "length_scales has been removed since models are now dimensional"
        )

    @length_scales.setter
    def length_scales(self, values):
        raise NotImplementedError(
            "length_scales has been removed since models are now dimensional"
        )

    @property
    def geometry(self):
        """Returns the geometry of the model."""
        return self._geometry

    @property
    def default_var_pts(self):
        """Returns a dictionary of the default variable points for the model, which is empty by default."""
        return {}

    @property
    def default_geometry(self):
        """Returns a  dictionary of the default geometry for the model, which is empty by default."""
        return {}

    @property
    def default_submesh_types(self):
        """Returns a dictionary of the default submesh types for the model, which is empty by default."""
        return {}

    @property
    def default_spatial_methods(self):
        """Returns a dictionary of the default spatial methods for the model, which is empty by default."""
        return {}

    @property
    def default_solver(self):
        """Returns the default solver for the model, based on whether it is an ODE/DAE or algebraic model."""
        if len(self.rhs) == 0 and len(self.algebraic) != 0:
            return pybamm.CasadiAlgebraicSolver()
        else:
            return pybamm.IDAKLUSolver()

    @property
    def default_quick_plot_variables(self):
        """Returns the default variables for quick plotting (None by default)."""
        return None

    @property
    def default_parameter_values(self):
        """Returns the default parameter values for the model (an empty set of parameters by default)."""
        return pybamm.ParameterValues({})

    @property
    def parameters(self):
        """Returns a  list of all parameter symbols used in the model."""
        self._parameters = self._find_symbols(
            (pybamm.Parameter, pybamm.InputParameter, pybamm.FunctionParameter)
        )
        return self._parameters

    @property
    def input_parameters(self):
        """Returns a list of all input parameter symbols used in the model."""
        if self._input_parameters is None:
            self._input_parameters = self._find_symbols(pybamm.InputParameter)
        return self._input_parameters

    @property
    def is_standard_form_dae(self):
        """
        Check if the model is a DAE in standard form with a mass matrix that is all
        zeros except for along the diagonal, which is either ones or zeros.
        """
        if self._is_standard_form_dae is None:
            self._is_standard_form_dae = self._check_standard_form_dae()
        return self._is_standard_form_dae

    @property
    def calc_esoh(self):
        """Whether to include eSOH variables in the summary variables."""
        return self._calc_esoh

    @property
    def algebraic_root_solver(self):
        return self._algebraic_root_solver

    @algebraic_root_solver.setter
    def algebraic_root_solver(self, algebraic_root_solver):
        self._algebraic_root_solver = algebraic_root_solver

    def get_parameter_info(self, by_submodel=False):
        """
        Extracts the parameter information and returns it as a dictionary.
        To get a list of all parameter-like objects without extra information,
        use :py:attr:`model.parameters`.

        Parameters
        ----------
        by_submodel : bool, optional
            Whether to return the parameter info sub-model wise or not (default False)
        """
        parameter_info = {}

        if by_submodel:
            for submodel_name, submodel_vars in self._variables_by_submodel.items():
                submodel_info = {}
                for var_name, var_symbol in submodel_vars.items():
                    if isinstance(var_symbol, pybamm.Parameter):
                        submodel_info[var_name] = (var_symbol, "Parameter")
                    elif isinstance(var_symbol, pybamm.InputParameter):
                        if not var_symbol.domain:
                            submodel_info[var_name] = (var_symbol, "InputParameter")
                        else:
                            submodel_info[var_name] = (
                                var_symbol,
                                f"InputParameter in {var_symbol.domain}",
                            )
                    elif isinstance(var_symbol, pybamm.FunctionParameter):
                        input_names = "', '".join(var_symbol.input_names)
                        submodel_info[var_name] = (
                            var_symbol,
                            f"FunctionParameter with inputs(s) '{input_names}'",
                        )
                    else:
                        submodel_info[var_name] = (var_symbol, "Unknown Type")

                parameters = self._find_symbols_by_submodel(
                    pybamm.Parameter, submodel_name
                )
                for param in parameters:
                    submodel_info[param.name] = (param, "Parameter")

                input_parameters = self._find_symbols_by_submodel(
                    pybamm.InputParameter, submodel_name
                )
                for input_param in input_parameters:
                    if not input_param.domain:
                        submodel_info[input_param.name] = (
                            input_param,
                            "InputParameter",
                        )
                    else:
                        submodel_info[input_param.name] = (
                            input_param,
                            f"InputParameter in {input_param.domain}",
                        )

                function_parameters = self._find_symbols_by_submodel(
                    pybamm.FunctionParameter, submodel_name
                )
                for func_param in function_parameters:
                    if func_param.name not in parameter_info:
                        input_names = "', '".join(func_param.input_names)
                        submodel_info[func_param.name] = (
                            func_param,
                            f"FunctionParameter with inputs(s) '{input_names}'",
                        )

                parameter_info[submodel_name] = submodel_info

        else:
            parameters = self._find_symbols(pybamm.Parameter)
            for param in parameters:
                parameter_info[param.name] = (param, "Parameter")

            input_parameters = self._find_symbols(pybamm.InputParameter)
            for input_param in input_parameters:
                if not input_param.domain:
                    parameter_info[input_param.name] = (input_param, "InputParameter")
                else:
                    parameter_info[input_param.name] = (
                        input_param,
                        f"InputParameter in {input_param.domain}",
                    )

            function_parameters = self._find_symbols(pybamm.FunctionParameter)
            for func_param in function_parameters:
                if func_param.name not in parameter_info:
                    input_names = "', '".join(func_param.input_names)
                    parameter_info[func_param.name] = (
                        func_param,
                        f"FunctionParameter with inputs(s) '{input_names}'",
                    )

        return parameter_info

    def _calculate_max_lengths(self, parameter_dict):
        """
        Calculate the maximum length of parameters and parameter type in a dictionary

        Parameters
        ----------
        parameter_dict : dict
            The dict from which maximum lengths are calculated
        """
        max_name_length = max(
            len(getattr(parameter, "name", str(parameter)))
            for parameter, _ in parameter_dict.values()
        )
        max_type_length = max(
            len(parameter_type) for _, parameter_type in parameter_dict.values()
        )

        return max_name_length, max_type_length

    def _check_standard_form_dae(self):
        """
        Check if the model is a DAE in standard form with a mass matrix that is all
        zeros except for along the diagonal, which is either ones or zeros.

        For example, the following is standard form:
        M*y' = f(y, y', t)

        M = [I 0
             0 0]

        The following explicit ODE is also a standard form DAE:
        M = I

        The following is not standard form:
        M = [2I 0
             0 0]
        """
        if self.mass_matrix is None or self.mass_matrix_inv is None:
            return False
        # Check that the mass matrix inverse is an identity matrix
        mass_matrix_inv = self.mass_matrix_inv.entries
        if scipy.sparse.issparse(mass_matrix_inv):
            identity = scipy.sparse.identity(mass_matrix_inv.shape[0])
            return (mass_matrix_inv - identity).nnz == 0
        else:
            return np.allclose(mass_matrix_inv, np.eye(mass_matrix_inv.shape[0]))

    def _format_table_row(
        self, param_name, param_type, max_name_length, max_type_length
    ):
        """
        Format the parameter information in a formatted table

        Parameters
        ----------
        param_name : str
            The name of the parameter
        param_type : str
            The type of the parameter
        max_name_length : int
            The maximum length of the parameter in the dictionary
        max_type_length : int
            The maximum length of the parameter type in the dictionary
        """
        param_name_lines = [
            param_name[i : i + max_name_length]
            for i in range(0, len(param_name), max_name_length)
        ]
        param_type_lines = [
            param_type[i : i + max_type_length]
            for i in range(0, len(param_type), max_type_length)
        ]
        max_lines = max(len(param_name_lines), len(param_type_lines))

        return [
            f"│ {param_name_lines[i]:<{max_name_length}} │ {param_type_lines[i]:<{max_type_length}} │"
            for i in range(max_lines)
        ]

    def print_parameter_info(self, by_submodel=False):
        """
        Print parameter information in a formatted table from a dictionary of parameters

        Parameters
        ----------
        by_submodel : bool, optional
            Whether to print the parameter info sub-model wise or not (default False)
        """

        if by_submodel:
            parameter_info = self.get_parameter_info(by_submodel=True)
            for submodel_name, submodel_vars in parameter_info.items():
                if not submodel_vars:
                    print(f"'{submodel_name}' submodel parameters: \nNo parameters\n")
                else:
                    print(f"'{submodel_name}' submodel parameters:")
                    (
                        max_param_name_length,
                        max_param_type_length,
                    ) = self._calculate_max_lengths(submodel_vars)

                    table = [
                        f"┌─{'─' * max_param_name_length}─┬─{'─' * max_param_type_length}─┐",
                        f"│ {'Parameter':<{max_param_name_length}} │ {'Type of parameter':<{max_param_type_length}} │",
                        f"├─{'─' * max_param_name_length}─┼─{'─' * max_param_type_length}─┤",
                    ]

                    for param, param_type in submodel_vars.values():
                        param_name = getattr(param, "name", str(param))
                        table.extend(
                            self._format_table_row(
                                param_name,
                                param_type,
                                max_param_name_length,
                                max_param_type_length,
                            )
                        )
                    table.extend(
                        [
                            f"└─{'─' * max_param_name_length}─┴─{'─' * max_param_type_length}─┘",
                        ]
                    )
                    table = "\n".join(table) + "\n"
                    table.encode("utf-8")
                    print(table)

        else:
            info = self.get_parameter_info()
            max_param_name_length, max_param_type_length = self._calculate_max_lengths(
                info
            )

            table = [
                f"┌─{'─' * max_param_name_length}─┬─{'─' * max_param_type_length}─┐",
                f"│ {'Parameter':<{max_param_name_length}} │ {'Type of parameter':<{max_param_type_length}} │",
                f"├─{'─' * max_param_name_length}─┼─{'─' * max_param_type_length}─┤",
            ]

            for param, param_type in info.values():
                param_name = getattr(param, "name", str(param))
                table.extend(
                    self._format_table_row(
                        param_name,
                        param_type,
                        max_param_name_length,
                        max_param_type_length,
                    )
                )

            table.extend(
                [
                    f"└─{'─' * max_param_name_length}─┴─{'─' * max_param_type_length}─┘",
                ]
            )

            table = "\n".join(table) + "\n"
            table.encode("utf-8")
            print(table)

    def _find_symbols(self, typ):
        """Find all the instances of `typ` in the model"""
        unpacker = pybamm.SymbolUnpacker(typ)
        all_input_parameters = unpacker.unpack_list_of_symbols(
            list(self.rhs.values())
            + list(self.algebraic.values())
            + list(self.initial_conditions.values())
            + [
                x[side][0]
                for x in self.boundary_conditions.values()
                for side in x.keys()
            ]
            + list(self.variables.values())
            + [event.expression for event in self.events]
        )
        return list(all_input_parameters)

    def _find_symbols_by_submodel(self, typ, submodel):
        """Find all the instances of `typ` in the submodel"""
        unpacker = pybamm.SymbolUnpacker(typ)
        all_input_parameters = unpacker.unpack_list_of_symbols(
            list(self.submodels[submodel].rhs.values())
            + list(self.submodels[submodel].algebraic.values())
            + list(self.submodels[submodel].initial_conditions.values())
            + [
                x[side][0]
                for x in self.submodels[submodel].boundary_conditions.values()
                for side in x.keys()
            ]
            + list(self._variables_by_submodel[submodel].values())
            + [event.expression for event in self.submodels[submodel].events]
        )
        return list(all_input_parameters)

    def new_copy(self):
        """
        Creates a copy of the model, explicitly copying all the mutable attributes
        to avoid issues with shared objects.
        """
        new_model = copy.copy(self)
        new_model._rhs = self.rhs.copy()
        new_model._algebraic = self.algebraic.copy()
        new_model._initial_conditions = self.initial_conditions.copy()
        new_model._boundary_conditions = self.boundary_conditions.copy()
        new_model._variables = self.variables.copy()
        new_model._events = self.events.copy()
        new_model._variables_casadi = self._variables_casadi.copy()
        return new_model

    def update(self, *submodels):
        """
        Update model to add new physics from submodels

        Parameters
        ----------
        submodel : iterable of :class:`pybamm.BaseModel`
            The submodels from which to create new model
        """
        for submodel in submodels:
            # check and then update dicts
            self.check_and_combine_dict(self._rhs, submodel.rhs)
            self.check_and_combine_dict(self._algebraic, submodel.algebraic)
            self.check_and_combine_dict(
                self._initial_conditions, submodel.initial_conditions
            )
            self.check_and_combine_dict(
                self._boundary_conditions, submodel.boundary_conditions
            )
            self.variables.update(submodel.variables)  # keys are strings so no check
            self._events += submodel.events

    def build_fundamental(self):
        # Get the fundamental variables
        self._variables_by_submodel = {submodel: {} for submodel in self.submodels}
        for submodel_name, submodel in self.submodels.items():
            pybamm.logger.debug(
                f"Getting fundamental variables for {submodel_name} submodel ({self.name})"
            )
            submodel_fundamental_variables = submodel.get_fundamental_variables()
            self._variables_by_submodel[submodel_name].update(
                submodel_fundamental_variables
            )
            self.variables.update(submodel_fundamental_variables)

        self._built_fundamental = True

    def build_coupled_variables(self):
        # Note: pybamm will try to get the coupled variables for the submodels in the
        # order they are set by the user. If this fails for a particular submodel,
        # return to it later and try again. If setting coupled variables fails and
        # there are no more submodels to try, raise an error.
        submodels = list(self.submodels.keys())
        count = 0
        # For this part the FuzzyDict of variables is briefly converted back into a
        # normal dictionary for speed with KeyErrors
        self._variables = dict(self._variables)
        while len(submodels) > 0:
            count += 1
            for submodel_name, submodel in self.submodels.items():
                if submodel_name in submodels:
                    pybamm.logger.debug(
                        f"Getting coupled variables for {submodel_name} submodel ({self.name})"
                    )
                    try:
                        model_var_copy = self.variables.copy()
                        updated_variables = submodel.get_coupled_variables(
                            self.variables
                        )
                        self._variables_by_submodel[submodel_name].update(
                            {
                                key: updated_variables[key]
                                for key in updated_variables
                                if key not in model_var_copy
                            }
                        )
                        self.variables.update(updated_variables)
                        submodels.remove(submodel_name)
                    except KeyError as key:
                        if len(submodels) == 1 or count == 100:
                            # no more submodels to try
                            raise pybamm.ModelError(
                                f"Missing variable for submodel '{submodel_name}': {key}.\n"
                                + "Check the selected "
                                "submodels provide all of the required variables."
                            ) from key
                        else:
                            # try setting coupled variables on next loop through
                            pybamm.logger.debug(
                                f"Can't find {key}, trying other submodels first"
                            )
        # Convert variables back into FuzzyDict
        self.variables = pybamm.FuzzyDict(self._variables)

    def build_model_equations(self):
        # Set model equations
        for submodel_name, submodel in self.submodels.items():
            pybamm.logger.verbose(
                f"Setting rhs for {submodel_name} submodel ({self.name})"
            )

            submodel.set_rhs(self.variables)
            pybamm.logger.verbose(
                f"Setting algebraic for {submodel_name} submodel ({self.name})"
            )

            submodel.set_algebraic(self.variables)
            pybamm.logger.verbose(
                f"Setting boundary conditions for {submodel_name} submodel ({self.name})"
            )

            submodel.set_boundary_conditions(self.variables)
            pybamm.logger.verbose(
                f"Setting initial conditions for {submodel_name} submodel ({self.name})"
            )
            submodel.set_initial_conditions(self.variables)
            submodel.add_events_from(self.variables)
            pybamm.logger.verbose(f"Updating {submodel_name} submodel ({self.name})")
            self.update(submodel)
            self.check_no_repeated_keys()

    def build_model(self):
        self._build_model()
        self._built = True
        pybamm.logger.info(f"Finish building {self.name}")

    def _build_model(self):
        # Check if already built
        if self._built:
            raise pybamm.ModelError(
                """Model already built. If you are adding a new submodel, try using
                `model.update` instead."""
            )

        pybamm.logger.info(f"Start building {self.name}")

        if self._built_fundamental is False:
            self.build_fundamental()

        self.build_coupled_variables()

        self.build_model_equations()

    def set_initial_conditions_from(
        self,
        solution,
        inputs=None,
        inplace=True,
        return_type="model",
        mesh=None,
    ):
        """
        Update initial conditions with the final states from a Solution object or from
        a dictionary.
        This assumes that, for each variable in self.initial_conditions, there is a
        corresponding variable in the solution with the same name and size.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`, or dict
            The solution to use to initialize the model
        inputs : dict
            The dictionary of model input parameters.
        inplace : bool, optional
            Whether to modify the model inplace or create a new model (default True)
        return_type : str, optional
            Whether to return the model (default) or initial conditions ("ics")
        mesh : :class:`pybamm.Mesh`, optional
            The mesh to use to initialize the model
        """
        mesh = mesh or {}
        initial_conditions = {}
        is_dict_input = isinstance(solution, dict)
        if isinstance(solution, pybamm.Solution):
            solution = solution.last_state

        def _find_matching_variable(var, solution_model):
            if not (
                solution_model.is_discretised and solution_model.y_slices is not None
            ):
                return None
            var_id = var.id
            for sol_var in solution_model.y_slices.keys():
                if sol_var.id == var_id:
                    return sol_var
            return None

        def _extract_from_y_slices(var, solution_var, solution_model, solution):
            solution_y_slice = solution_model.y_slices[solution_var][0]
            y_last = solution.y[:, -1] if solution.y.ndim > 1 else solution.y

            # Validate slice bounds
            slice_stop = (
                solution_y_slice.stop
                if solution_y_slice.stop is not None
                else len(y_last)
            )
            slice_start = (
                solution_y_slice.start if solution_y_slice.start is not None else 0
            )
            if slice_start < 0 or slice_stop > len(y_last):
                return None

            # Extract scaled state vector values
            y_scaled = np.array(y_last[solution_y_slice])

            # Convert from scaled state vector to physical values
            # physical = reference + scale * y_scaled
            try:
                solution_scale = np.asarray(
                    solution_var.scale.evaluate()
                ) * np.ones_like(y_scaled)
                solution_reference = np.asarray(
                    solution_var.reference.evaluate()
                ) * np.ones_like(y_scaled)
                return solution_reference + solution_scale * y_scaled
            except (TypeError, ValueError, AttributeError):
                # Fall back to dict lookup
                return None

        def _extract_final_time_step(var_data):
            var_data = np.array(var_data)
            if var_data.ndim == 0:
                return var_data
            elif var_data.ndim == 1:
                return np.array(var_data[-1:])
            elif var_data.ndim == 2:
                return np.array(var_data[:, -1])
            elif var_data.ndim == 3:
                return var_data[:, :, -1].flatten(order="F")
            elif var_data.ndim == 4:
                return var_data[:, :, :, -1].flatten(order="F")
            else:
                raise NotImplementedError("Variable must be 0D, 1D, 2D, 3D, or 4D")

        def get_final_state_eval(final_state):
            # If it's a ProcessedVariable, extract .data first
            if isinstance(solution, pybamm.Solution) and hasattr(final_state, "data"):
                final_state = final_state.data

            final_state = np.array(final_state)

            # Extract final state from time series
            if final_state.ndim == 0:
                return np.array([final_state])
            elif final_state.ndim == 1:
                # 1D arrays are already final state (from y_slices or processed from dict)
                return np.array(final_state)
            elif final_state.ndim == 2:
                return np.array(final_state[:, -1])
            elif final_state.ndim == 3:
                return final_state[:, :, -1].flatten(order="F")
            elif final_state.ndim == 4:
                return final_state[:, :, :, -1].flatten(order="F")
            else:
                raise NotImplementedError("Variable must be 0D, 1D, 2D, 3D, or 4D")

        def get_variable_state(var):
            var_name = var.name

            # Try y_slices for discretised models
            if (
                self.is_discretised
                and var in self.y_slices
                and isinstance(solution, pybamm.Solution)
                and len(solution.all_models) > 0
            ):
                try:
                    solution_model = solution.all_models[-1]
                    solution_var = _find_matching_variable(var, solution_model)
                    if solution_var is not None:
                        final_state = _extract_from_y_slices(
                            var, solution_var, solution_model, solution
                        )
                        if final_state is not None:
                            return final_state
                except (KeyError, AttributeError, IndexError, TypeError):
                    pass

            # Fall back to solution[var_name] lookup
            try:
                var_data = solution[var_name]
                # For dict inputs, extract final time step here
                if is_dict_input:
                    return _extract_final_time_step(var_data)
                else:
                    return var_data
            except KeyError as e:
                raise pybamm.ModelError(
                    "To update a model from a solution, each variable in "
                    "model.initial_conditions must appear in the solution with "
                    "the same key as the variable name. In the solution provided, "
                    f"'{e.args[0]}' was not found."
                ) from e

        for var in self.initial_conditions:
            if isinstance(var, pybamm.Variable) or isinstance(
                var, pybamm.Concatenation
            ):
                try:
                    final_state = get_variable_state(var)
                    final_state_eval = get_final_state_eval(final_state)
                except pybamm.ModelError as e:
                    if isinstance(var, pybamm.Concatenation):
                        children = []
                        for child in var.orphans:
                            final_state = get_variable_state(child)
                            final_state_eval = get_final_state_eval(final_state)
                            children.append(final_state_eval)
                        final_state_eval = np.concatenate(children)
                    else:
                        raise e
            else:
                raise NotImplementedError(
                    "Variable must have type 'Variable' or 'Concatenation'"
                )

            # If the model is already discretised, then the initial conditions must
            # be scaled and offset (otherwise, this is done when the model is
            # discretised)
            if self.is_discretised:
                scale, reference = var.scale, var.reference
            else:
                scale, reference = pybamm.Scalar(1), pybamm.Scalar(0)
            initial_conditions[var] = (
                pybamm.Vector(final_state_eval) - reference
            ) / scale.evaluate(inputs=inputs)

        # Also update the concatenated initial conditions if the model is already
        # discretised
        if self.is_discretised:
            # Unpack slices for sorting
            y_slices = {var: slce for var, slce in self.y_slices.items()}
            slices = []
            for symbol in self.initial_conditions.keys():
                if isinstance(symbol, pybamm.Concatenation):
                    # must append the slice for the whole concatenation, so that
                    # equations get sorted correctly
                    slices.append(
                        slice(
                            y_slices[symbol.children[0]][0].start,
                            y_slices[symbol.children[-1]][0].stop,
                        )
                    )
                else:
                    slices.append(y_slices[symbol][0])
            equations = list(initial_conditions.values())
            # sort equations according to slices
            sorted_equations = [
                eq for _, eq in sorted(zip(slices, equations, strict=False))
            ]
            concatenated_initial_conditions = pybamm.NumpyConcatenation(
                *sorted_equations
            )
        else:
            concatenated_initial_conditions = None

        if return_type == "model":
            if inplace is True:
                model = self
            else:
                model = self.new_copy()

            model.initial_conditions = initial_conditions
            model.concatenated_initial_conditions = concatenated_initial_conditions
            return model
        elif return_type == "ics":
            return initial_conditions, concatenated_initial_conditions

    def check_and_combine_dict(self, dict1, dict2):
        # check that the key ids are distinct
        ids1 = set(x for x in dict1.keys())
        ids2 = set(x for x in dict2.keys())
        if len(ids1.intersection(ids2)) != 0:
            variables = ids1.intersection(ids2)
            raise pybamm.ModelError(
                f"Submodel incompatible: duplicate variables '{variables}'"
            )
        dict1.update(dict2)

    def check_well_posedness(self, post_discretisation=False):
        """
        Check that the model is well-posed by executing the following tests:
        - Model is not over- or underdetermined, by comparing keys and equations in rhs
        and algebraic. Overdetermined if more equations than variables, underdetermined
        if more variables than equations.
        - There is an initial condition in self.initial_conditions for each
        variable/equation pair in self.rhs
        - There are appropriate boundary conditions in self.boundary_conditions for each
        variable/equation pair in self.rhs and self.algebraic

        Parameters
        ----------
        post_discretisation : boolean
            A flag indicating tests to be skipped after discretisation
        """
        self.check_for_time_derivatives()
        self.check_well_determined(post_discretisation)
        self.check_algebraic_equations(post_discretisation)
        self.check_ics_bcs()
        self.check_no_repeated_keys()
        # Can't check variables after discretising, since Variable objects get replaced
        # by StateVector objects
        # Checking variables is slow, so only do it in debug mode
        if pybamm.settings.debug_mode is True and post_discretisation is False:
            self.check_variables()

    def check_for_time_derivatives(self):
        # Check that no variable time derivatives exist in the rhs equations
        for key, eq in self.rhs.items():
            for node in eq.pre_order():
                if isinstance(node, pybamm.VariableDot):
                    raise pybamm.ModelError(
                        "time derivative of variable found "
                        f"({node}) in rhs equation {key}"
                    )
                if isinstance(node, pybamm.StateVectorDot):
                    raise pybamm.ModelError(
                        "time derivative of state vector found "
                        f"({node}) in rhs equation {key}"
                    )

        # Check that no variable time derivatives exist in the algebraic equations
        for key, eq in self.algebraic.items():
            for node in eq.pre_order():
                if isinstance(node, pybamm.VariableDot):
                    raise pybamm.ModelError(
                        f"time derivative of variable found ({node}) in algebraic"
                        f"equation {key}"
                    )
                if isinstance(node, pybamm.StateVectorDot):
                    raise pybamm.ModelError(
                        f"time derivative of state vector found ({node}) in algebraic"
                        f"equation {key}"
                    )

    def check_well_determined(self, post_discretisation):
        """Check that the model is not under- or over-determined."""
        # Equations (differential and algebraic)
        # Get all the variables from differential and algebraic equations
        all_vars_in_rhs_keys = set()
        all_vars_in_algebraic_keys = set()
        all_vars_in_eqns = set()
        # Get all variables ids from rhs and algebraic keys and equations, and
        # from boundary conditions
        # For equations we look through the whole expression tree.
        # "Variables" can be Concatenations so we also have to look in the whole
        # expression tree
        unpacker = pybamm.SymbolUnpacker((pybamm.Variable, pybamm.VariableDot))

        for var, eqn in self.rhs.items():
            # Find all variables and variabledot objects
            vars_in_rhs_keys = unpacker.unpack_symbol(var)
            vars_in_eqns = unpacker.unpack_symbol(eqn)

            # Look only for Variable (not VariableDot) in rhs keys
            all_vars_in_rhs_keys.update(
                [var for var in vars_in_rhs_keys if isinstance(var, pybamm.Variable)]
            )
            all_vars_in_eqns.update(vars_in_eqns)
        for var, eqn in self.algebraic.items():
            # Find all variables and variabledot objects
            vars_in_algebraic_keys = unpacker.unpack_symbol(var)
            vars_in_eqns = unpacker.unpack_symbol(eqn)

            # Store ids only
            # Look only for Variable (not VariableDot) in algebraic keys
            all_vars_in_algebraic_keys.update(
                [
                    var
                    for var in vars_in_algebraic_keys
                    if isinstance(var, pybamm.Variable)
                ]
            )
            all_vars_in_eqns.update(vars_in_eqns)
        for _, side_eqn in self.boundary_conditions.items():
            for _, (eqn, _) in side_eqn.items():
                vars_in_eqns = unpacker.unpack_symbol(eqn)
                all_vars_in_eqns.update(vars_in_eqns)

        # If any keys are repeated between rhs and algebraic then the model is
        # overdetermined
        if not set(all_vars_in_rhs_keys).isdisjoint(all_vars_in_algebraic_keys):
            raise pybamm.ModelError("model is overdetermined (repeated keys)")
        # If any algebraic keys don't appear in the eqns (or bcs) then the model is
        # overdetermined (but rhs keys can be absent from the eqns, e.g. dcdt = -1 is
        # fine)
        # Skip this step after discretisation, as any variables in the equations will
        # have been discretised to slices but keys will still be variables
        extra_algebraic_keys = all_vars_in_algebraic_keys.difference(all_vars_in_eqns)
        if extra_algebraic_keys and not post_discretisation:
            raise pybamm.ModelError("model is overdetermined (extra algebraic keys)")
        # If any variables in the equations don't appear in the keys then the model is
        # underdetermined
        all_vars_in_keys = all_vars_in_rhs_keys.union(all_vars_in_algebraic_keys)
        extra_variables_in_equations = all_vars_in_eqns.difference(all_vars_in_keys)

        if extra_variables_in_equations:
            raise pybamm.ModelError("model is underdetermined (too many variables)")

    def check_algebraic_equations(self, post_discretisation):
        """
        Check that the algebraic equations are well-posed. After discretisation,
        there must be at least one StateVector in each algebraic equation.
        """
        if post_discretisation:
            # Check that each algebraic equation contains some StateVector
            for eqn in self.algebraic.values():
                if not eqn.has_symbol_of_classes(pybamm.StateVector):
                    raise pybamm.ModelError(
                        "each algebraic equation must contain at least one StateVector"
                    )
        else:
            # We do not perfom any checks before discretisation (most problematic
            # cases should be caught by `check_well_determined`)
            pass

    def check_ics_bcs(self):
        """Check that the initial and boundary conditions are well-posed."""
        # Initial conditions
        for var in self.rhs.keys():
            if var not in self.initial_conditions.keys():
                raise pybamm.ModelError(
                    f"no initial condition given for variable '{var}'"
                )

    def check_variables(self):
        # Create list of all Variable nodes that appear in the model's list of variables
        unpacker = pybamm.SymbolUnpacker(pybamm.Variable)
        all_vars = unpacker.unpack_list_of_symbols(self.variables.values())

        vars_in_keys = set()

        model_keys = list(self.rhs.keys()) + list(self.algebraic.keys())

        for var in model_keys:
            if isinstance(var, pybamm.Variable):
                vars_in_keys.add(var)
            # Key can be a concatenation
            elif isinstance(var, pybamm.Concatenation):
                vars_in_keys.update(var.children)

        for var in all_vars:
            if var not in vars_in_keys:
                raise pybamm.ModelError(
                    f"No key set for variable '{var}'. Make sure it is included in either "
                    "model.rhs or model.algebraic, in an unmodified form "
                    "(e.g. not Broadcasted)"
                )

    def check_no_repeated_keys(self):
        """Check that no equation keys are repeated."""
        rhs_keys = set(self.rhs.keys())
        alg_keys = set(self.algebraic.keys())

        if not rhs_keys.isdisjoint(alg_keys):
            raise pybamm.ModelError(
                f"Multiple equations specified for variables {rhs_keys.intersection(alg_keys)}"
            )

    def info(self, symbol_name):
        """
        Provides helpful summary information for a symbol.

        Parameters
        ----------
        symbol_name : str
        """
        # Should we deprecate this? Not really sure how it's used?

        div = "-----------------------------------------"
        symbol = find_symbol_in_model(self, symbol_name)

        if symbol is None:
            return None

        print(div)
        print(symbol_name, "\n")
        print(type(symbol))

        if isinstance(symbol, pybamm.FunctionParameter):
            print("")
            print("Inputs:")
            symbol.print_input_names()

        print(div)

    def check_discretised_or_discretise_inplace_if_0D(self):
        """
        Discretise model if it isn't already discretised
        This only works with purely 0D models, as otherwise the mesh and spatial
        method should be specified by the user
        """
        if self.is_discretised is False:
            try:
                disc = pybamm.Discretisation()
                disc.process_model(self)
            except pybamm.DiscretisationError as e:
                raise pybamm.DiscretisationError(
                    "Cannot automatically discretise model, model should be "
                    f"discretised before exporting casadi functions ({e})"
                ) from e

    def export_casadi_objects(self, variable_names, input_parameter_order=None):
        """
        Export the constituent parts of the model (rhs, algebraic, initial conditions,
        etc) as casadi objects.

        Parameters
        ----------
        variable_names : list
            Variables to be exported alongside the model structure
        input_parameter_order : list, optional
            Order in which the input parameters should be stacked.
            If input_parameter_order=None and len(self.input_parameters) > 1, a
            ValueError is raised (this helps to avoid accidentally using the wrong
            order)

        Returns
        -------
        casadi_dict : dict
            Dictionary of {str: casadi object} pairs representing the model in casadi
            format
        """
        self.check_discretised_or_discretise_inplace_if_0D()

        # Create casadi functions for the model
        t_casadi = casadi.MX.sym("t")
        y_diff = casadi.MX.sym("y_diff", self.concatenated_rhs.size)
        y_alg = casadi.MX.sym("y_alg", self.concatenated_algebraic.size)
        y_casadi = casadi.vertcat(y_diff, y_alg)

        # Read inputs
        inputs_wrong_order = {}
        for input_param in self.input_parameters:
            name = input_param.name
            inputs_wrong_order[name] = casadi.MX.sym(name, input_param._expected_size)
        # Sort according to input_parameter_order
        if input_parameter_order is None:
            if len(inputs_wrong_order) > 1:
                raise ValueError(
                    "input_parameter_order must be specified if there is more than one "
                    "input parameter"
                )
            inputs = inputs_wrong_order
        else:
            inputs = {name: inputs_wrong_order[name] for name in input_parameter_order}
        # Set up inputs
        inputs_stacked = casadi.vertcat(*[p for p in inputs.values()])

        # Convert initial conditions to casadi form
        y0 = self.concatenated_initial_conditions.to_casadi(
            t_casadi, y_casadi, inputs=inputs
        )
        x0 = y0[: self.concatenated_rhs.size]
        z0 = y0[self.concatenated_rhs.size :]

        # Convert rhs and algebraic to casadi form and calculate jacobians
        rhs = self.concatenated_rhs.to_casadi(t_casadi, y_casadi, inputs=inputs)
        jac_rhs = casadi.jacobian(rhs, y_casadi)
        algebraic = self.concatenated_algebraic.to_casadi(
            t_casadi, y_casadi, inputs=inputs
        )
        jac_algebraic = casadi.jacobian(algebraic, y_casadi)

        # For specified variables, convert to casadi
        variables = OrderedDict()
        for name in variable_names:
            var = self.variables[name]
            variables[name] = var.to_casadi(t_casadi, y_casadi, inputs=inputs)

        casadi_dict = {
            "t": t_casadi,
            "x": y_diff,
            "z": y_alg,
            "inputs": inputs_stacked,
            "rhs": rhs,
            "algebraic": algebraic,
            "jac_rhs": jac_rhs,
            "jac_algebraic": jac_algebraic,
            "variables": variables,
            "x0": x0,
            "z0": z0,
        }

        return casadi_dict

    def generate(
        self, filename, variable_names, input_parameter_order=None, cg_options=None
    ):
        """
        Generate the model in C, using CasADi.

        Parameters
        ----------
        filename : str
            Name of the file to which to save the code
        variable_names : list
            Variables to be exported alongside the model structure
        input_parameter_order : list, optional
            Order in which the input parameters should be stacked.
            If input_parameter_order=None and len(self.input_parameters) > 1, a
            ValueError is raised (this helps to avoid accidentally using the wrong
            order)
        cg_options : dict
            Options to pass to the code generator.
            See https://web.casadi.org/docs/#generating-c-code
        """
        model = self.export_casadi_objects(variable_names, input_parameter_order)

        # Read the exported objects
        t, x, z, p = model["t"], model["x"], model["z"], model["inputs"]
        x0, z0 = model["x0"], model["z0"]
        rhs, alg = model["rhs"], model["algebraic"]
        variables = model["variables"]
        jac_rhs, jac_alg = model["jac_rhs"], model["jac_algebraic"]

        # Create functions
        rhs_fn = casadi.Function("rhs_", [t, x, z, p], [rhs])
        alg_fn = casadi.Function("alg_", [t, x, z, p], [alg])
        jac_rhs_fn = casadi.Function("jac_rhs", [t, x, z, p], [jac_rhs])
        jac_alg_fn = casadi.Function("jac_alg", [t, x, z, p], [jac_alg])
        # Call these functions to initialize initial conditions
        # (initial conditions are not yet consistent at this stage)
        x0_fn = casadi.Function("x0", [p], [x0])
        z0_fn = casadi.Function("z0", [p], [z0])
        # Variables
        variables_stacked = casadi.vertcat(*variables.values())
        variables_fn = casadi.Function("variables", [t, x, z, p], [variables_stacked])

        # Write C files
        cg_options = cg_options or {}
        C = casadi.CodeGenerator(filename, cg_options)
        C.add(rhs_fn)
        C.add(alg_fn)
        C.add(jac_rhs_fn)
        C.add(jac_alg_fn)
        C.add(x0_fn)
        C.add(z0_fn)
        C.add(variables_fn)
        C.generate()

    def latexify(self, filename=None, newline=True, output_variables=None):
        """
        Converts all model equations in latex.

        Parameters
        ----------
        filename: str (optional)
            Accepted file formats - any image format, pdf and tex
            Default is None, When None returns all model equations in latex
            If not None, returns all model equations in given file format.

        newline: bool (optional)
            Default is True, If True, returns every equation in a new line.
            If False, returns the list of all the equations.

        Load model
        >>> model = pybamm.lithium_ion.SPM()

        This will returns all model equations in png
        >>> model.latexify("equations.png") # doctest: +SKIP

        This will return all the model equations in latex
        >>> model.latexify() # doctest: +SKIP

        This will return the list of all the model equations
        >>> model.latexify(newline=False) # doctest: +SKIP

        This will return first five model equations
        >>> model.latexify(newline=False)[1:5] # doctest: +SKIP
        """
        from pybamm.expression_tree.operations.latexify import Latexify

        return Latexify(self, filename, newline).latexify(
            output_variables=output_variables
        )

    def process_parameters_and_discretise(self, symbol, parameter_values, disc):
        """
        Process parameters and discretise a symbol using supplied parameter values
        and discretisation. Note: care should be taken if using spatial operators
        on dimensional symbols. Operators in pybamm are written in non-dimensional
        form, so may need to be scaled by the appropriate length scale. It is
        recommended to use this method on non-dimensional symbols.

        Parameters
        ----------
        symbol : :class:`pybamm.Symbol`
            Symbol to be processed
        parameter_values : :class:`pybamm.ParameterValues`
            The parameter values to use during processing
        disc : :class:`pybamm.Discretisation`
            The discrisation to use

        Returns
        -------
        :class:`pybamm.Symbol`
            Processed symbol
        """
        # Set y slices
        if disc.y_slices == {}:
            variables = list(self.rhs.keys()) + list(self.algebraic.keys())
            for variable in variables:
                variable.bounds = tuple(
                    [
                        parameter_values.process_symbol(bound)
                        for bound in variable.bounds
                    ]
                )
            disc.set_variable_slices(variables)

        # Set boundary conditions (also requires setting parameter values)
        if disc.bcs == {}:
            self.boundary_conditions = parameter_values.process_boundary_conditions(
                self
            )
            disc.bcs = disc.process_boundary_conditions(self)

        # Process
        param_symbol = parameter_values.process_symbol(symbol)
        disc_symbol = disc.process_symbol(param_symbol)

        return disc_symbol

    def save_model(self, filename=None, mesh=None, variables=None):
        """
        Write out a discretised model to a JSON file

        Parameters
        ----------
        filename: str, optional
        The desired name of the JSON file. If no name is provided, one will be created
        based on the model name, and the current datetime.
        """
        if variables and not mesh:
            warnings.warn(
                """
                Serialisation: Variables are being saved without a mesh.
                Plotting may not be available.
                """,
                pybamm.ModelWarning,
                stacklevel=2,
            )

        Serialise().save_model(self, filename=filename, mesh=mesh, variables=variables)


def load_model(filename, battery_model: BaseModel | None = None):
    """
    Load in a saved model from a JSON file

    Parameters
    ----------
    filename: str
        Path to the JSON file containing the serialised model file
    battery_model: :class: pybamm.BaseBatteryModel, optional
            PyBaMM model to be created (e.g. pybamm.lithium_ion.SPM), which will
            override any model names within the file. If None, the function will look
            for the saved object path, present if the original model came from PyBaMM.
    """
    return Serialise().load_model(filename, battery_model)


# helper functions for finding symbols
def find_symbol_in_tree(tree, name):
    if name == tree.name:
        return tree
    elif len(tree.children) > 0:
        for child in tree.children:
            child_return = find_symbol_in_tree(child, name)
            if child_return is not None:
                return child_return


def find_symbol_in_dict(dic, name):
    for tree in dic.values():
        tree_return = find_symbol_in_tree(tree, name)
        if tree_return is not None:
            return tree_return


def find_symbol_in_model(model, name):
    dics = [model.rhs, model.algebraic, model.variables]
    for dic in dics:
        dic_return = find_symbol_in_dict(dic, name)
        if dic_return is not None:
            return dic_return


class EquationDict(dict):
    def __init__(self, name, equations):
        self.name = name
        equations = self.check_and_convert_equations(equations)
        super().__init__(equations)

    def __setitem__(self, key, value):
        """Call the update functionality when doing a setitem."""
        self.update({key: value})

    def update(self, equations):
        equations = self.check_and_convert_equations(equations)
        super().update(equations)

    def check_and_convert_equations(self, equations):
        """
        Convert any scalar equations in dict to 'pybamm.Scalar'
        and check that domains are consistent
        """
        # Convert any numbers to a pybamm.Scalar
        for var, eqn in equations.items():
            if isinstance(eqn, numbers.Number):
                eqn = pybamm.Scalar(eqn)
                equations[var] = eqn
            if not (var.domain == eqn.domain or var.domain == [] or eqn.domain == []):
                raise pybamm.DomainError(
                    f"variable and equation in '{self.name}' must have the same domain"
                )

        # For initial conditions, check that the equation doesn't contain any
        # Variable objects
        # skip this if the dictionary has no "name" attribute (which will be the case
        # after pickling)
        if hasattr(self, "name") and self.name == "initial_conditions":
            for var, eqn in equations.items():
                if eqn.has_symbol_of_classes(pybamm.Variable):
                    unpacker = pybamm.SymbolUnpacker(pybamm.Variable)
                    variable_in_equation = next(iter(unpacker.unpack_symbol(eqn)))
                    raise TypeError(
                        "Initial conditions cannot contain 'Variable' objects, "
                        f"but '{variable_in_equation!r}' found in initial conditions for '{var}'"
                    )

        return equations


class BoundaryConditionsDict(dict):
    def __init__(self, bcs):
        bcs = self.check_and_convert_bcs(bcs)
        super().__init__(bcs)

    def __setitem__(self, key, value):
        """Call the update functionality when doing a setitem."""
        self.update({key: value})

    def update(self, bcs):
        bcs = self.check_and_convert_bcs(bcs)
        super().update(bcs)

    def check_and_convert_bcs(self, boundary_conditions):
        """Convert any scalar bcs in dict to 'pybamm.Scalar', and check types."""
        # Convert any numbers to a pybamm.Scalar
        for var, bcs in boundary_conditions.items():
            for side, bc in bcs.items():
                if isinstance(bc[0], numbers.Number):
                    # typ is the type of the bc, e.g. "Dirichlet" or "Neumann"
                    eqn, typ = boundary_conditions[var][side]
                    boundary_conditions[var][side] = (pybamm.Scalar(eqn), typ)
                # Check types
                if bc[1] not in ["Dirichlet", "Neumann"]:
                    raise pybamm.ModelError(
                        f"boundary condition types must be Dirichlet or Neumann, not '{bc[1]}'"
                    )

        return boundary_conditions



import pathlib

import pooch


class DataLoader:
    """
    Data Loader class for downloading and loading data files upstream at https://github.com/pybamm-team/pybamm-data/

    The following files are listed in the registry -

    COMSOL Results
    ---------------

    :footcite:t:`Andersson2019`
    :footcite:t:`Doyle1993`
    :footcite:t:`Harris2020`
    :footcite:t:`Marquis2019`
    :footcite:t:`Marquis2020`

    - comsol_01C.json
    - comsol_05C.json
    - comsol_1C.json
    - comsol_1plus1D_3C.json
    - comsol_2C.json
    - comsol_3C.json

    Kokam SLPB 75106100 discharge data from Ecker et al (2015)
    ----------------------------------------------------------

    :footcite:t:`Ecker2015i`
    :footcite:t:`Ecker2015ii`

    - Ecker_1C.csv
    - Ecker_5C.csv

    Enertech cells - discharge results for beginning of life
    --------------------------------------------------------

    :footcite:t:`Andersson2019`
    :footcite:t:`Doyle1993`
    :footcite:t:`Harris2020`
    :footcite:t:`Marquis2019`
    :footcite:t:`Ai2019`
    :footcite:t:`Deshpande2012`
    :footcite:t:`Timms2021`

    - 0.1C_discharge_U.txt
    - 0.1C_discharge_displacement.txt
    - 0.5C_discharge_T.txt
    - 0.5C_discharge_U.txt
    - 0.5C_discharge_displacement.txt
    - 1C_discharge_T.txt
    - 1C_discharge_U.txt
    - 1C_discharge_displacement.txt
    - 2C_discharge_T.txt
    - 2C_discharge_U.txt
    - 2C_discharge_displacement.txt
    - stn_2C.txt
    - stp_2C.txt


    Drive cycles
    ------------

    :footcite:t:`Andersson2019`
    :footcite:t:`Doyle1993`
    :footcite:t:`Harris2020`
    :footcite:t:`Marquis2019`
    :footcite:t:`Marquis2020`

    - UDDS.csv
    - US06.csv
    - WLTC.csv
    - car_current.csv


    """

    def __init__(self):
        """
        Create a pooch registry with the following data files available upstream at https://github.com/pybamm-team/pybamm-data/
        """
        self.version = "v1.0.1"  # Version of pybamm-data release
        self.path = pooch.os_cache("pybamm")
        self.files = {
            # COMSOL results
            "comsol_01C.json": "sha256:bc5136fe961e269453bdc31fcaa97376d6f8c347d570fd30ce4b7660c68ae22c",
            "comsol_05C.json": "sha256:3b044135ad88bdb88959304a33fe42b654d5ef7ef79d1271dd909cec55b257fb",
            "comsol_1C.json": "sha256:d45e3ab482c497c37ebbc68898da22bab0b0263992d8f2302502028bfd5ba0e9",
            "comsol_1plus1D_3C.json": "sha256:cdd5759202f9c7887d2ea6032f82212f2ca89297191fe5282b8812e1a09b1e1f",
            "comsol_2C.json": "sha256:15c2637f54bf1639621c58795db859cb08611c8182b7b20ade10e4c3e2839a5b",
            "comsol_3C.json": "sha256:11d5afccb70be85d4ac7e61d413c6e0f5e318e1635b1347c9a3c6784119711e6",
            # Kokam SLPB 75106100 discharge data from Ecker et al (2015)
            "Ecker_1C.csv": "sha256:428dc5113a6430492f430fb9e895f67d3e20f5643dc49a1cc0a922b92a5a8e01",
            "Ecker_5C.csv": "sha256:a89f8bf6e305b2a4195e1fae5e803277a40ed7557d263ef726f621803dcbb495",
            # Enertech cells - discharge results for beginning of life
            "0.1C_discharge_U.txt": "sha256:7b9fcd137441eea4ab686faee8d57fe242c5544400939ef358ccd99c63c9579d",
            "0.1C_discharge_displacement.txt": "sha256:f1329731ead5a82a2be9851cf80e4c6d68dd0774e07aee5361e2af3ab420d7be",
            "0.5C_discharge_T.txt": "sha256:2140b2f6bd698135d09a25b1f04c271d35a3a02999ace118b10389e01defa2ae",
            "0.5C_discharge_U.txt": "sha256:9ed8368b2c6149d2a69218e7df6aaade2511c9f7f6fc7932cda153d9a3a10f39",
            "0.5C_discharge_displacement.txt": "sha256:8098565ff99bc938864797b402f483c1c64a583d6db85d086f39ab0e7b638dd1",
            "1C_discharge_T.txt": "sha256:97308dfd7f7dd6c434e30f6c00fb6707c43c963855bb0800e0336809d5cc3756",
            "1C_discharge_U.txt": "sha256:8fc19de45172215d65c56522c224e6fc700ee443db236b814238a829b7a14c3a",
            "1C_discharge_displacement.txt": "sha256:c2e8617ac48a20921da1b40bbebac479a0a143edf16b12b2e1ff9aaaf1a32ff4",
            "2C_discharge_T.txt": "sha256:4bd688fb7653539701fe3df61857474b4d54e8b142c84fdc4c8b92b9573fa5d0",
            "2C_discharge_U.txt": "sha256:7b3c24b5e6df377075002abc2f62bab7c88b27d826812ba5a4c8385a1a12e723",
            "2C_discharge_displacement.txt": "sha256:2b11513d80827c762325c819a084b87b3a239af7d112f234c9871481760a0013",
            "stn_2C.txt": "sha256:bb2f90ccfd2cd86ad589287caae13470e554df2f4f47f0f583a5a7e3e6bd9d4c",
            "stp_2C.txt": "sha256:6fe73b3a18e5fcfb95151dfd7d34c3cbe929792631447ed3ec88c047c9778223",
            # Drive cycles
            "UDDS.csv": "sha256:9fe6558c17aad3cc08109186923aeb7459cd3097a381c44e854bf22dd12a5a2a",
            "US06.csv": "sha256:5909eb2ec7983fae86a050ff3b35a2041d0ab698710a6b0f95d5816e348077ba",
            "WLTC.csv": "sha256:bb2f95018a44ac1425cb9c787c34721192af502c7385f1358f28e4f75df11fd8",
            "car_current.csv": "sha256:4305b91b9df073cb048c25dd3fae725e06a94fe200e322e5c08db290d6799e36",
            # BPX files
            "nmc_pouch_cell_BPX.json": "sha256:27261e2a7012725ed16c0c8b799b870388bd6c77f86bbf1949353286569c7d0d",
        }
        self.registry = pooch.create(
            path=self.path,
            base_url=f"https://github.com/pybamm-team/pybamm-data/releases/download/{self.version}",
            version=self.version,
            registry=self.files,
        )

    def get_data(self, filename: str):
        """
        Fetches the data file from upstream and stores it in the local cache directory under pybamm directory.

        Parameters
        ----------
        filename : str
            Name of the data file to be fetched from the registry.
        Returns
        -------
        pathlib.PurePath
        """
        self.registry.fetch(filename)
        return pathlib.Path(f"{self.path}/{self.version}/{filename}")

    def show_registry(self):
        """
        Prints the name of all the files present in the registry.

        Returns
        -------
        list
        """
        return list(self.files.keys())



#
# Base class for particle cracking models.
#
from __future__ import annotations

from typing import TYPE_CHECKING

import pybamm

if TYPE_CHECKING:
    from typing import Any


class BaseMechanics(pybamm.BaseSubModel):
    """
    Base class for particle mechanics models, referenced from :footcite:t:`Ai2019` and
    :footcite:t:`Deshpande2012`.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : dict, optional
        Dictionary of either the electrode for "positive" or "Negative"
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")

    """

    # Physical constants for stress/displacement calculations (Ai2019)
    STRESS_GEOMETRIC_FACTOR = 3.0
    DISPLACEMENT_GEOMETRIC_FACTOR = 3.0
    CRACK_ROUGHNESS_FACTOR = 2.0

    def __init__(
        self,
        param: Any,
        domain: str | dict[str, str],
        options: dict[str, Any],
        phase: str = "primary",
    ) -> None:
        self.size_distribution = options["particle size"] == "distribution"
        super().__init__(param, domain, options=options, phase=phase)

    def _build_var_name(
        self, quantity: str, unit: str = "", averaged: bool = False
    ) -> str:
        """
        Construct standardized variable names for mechanics quantities.

        Parameters
        ----------
        quantity : str
            The physical quantity (e.g., "particle crack length")
        unit : str, optional
            The unit of measurement (e.g., "[m]")
        averaged : bool, optional
            Whether this is an x-averaged quantity

        Returns
        -------
        str
            The formatted variable name
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_param.phase_name

        prefix = f"X-averaged {domain}" if averaged else Domain
        unit_str = f" {unit}" if unit else ""

        return f"{prefix} {phase_name}{quantity}{unit_str}"

    def _get_standard_variables(self, l_cr: pybamm.Symbol) -> dict[str, pybamm.Symbol]:
        """Generate standard crack length variables."""
        l_cr_av = pybamm.x_average(l_cr)
        return {
            self._build_var_name("particle crack length", "[m]"): l_cr,
            self._build_var_name(
                "particle crack length", "[m]", averaged=True
            ): l_cr_av,
        }

    def _get_standard_size_distribution_variables(
        self, l_cr_dist: pybamm.Symbol
    ) -> dict[str, pybamm.Symbol]:
        """Generate standard crack length distribution variables."""
        l_cr_av_dist = pybamm.x_average(l_cr_dist)
        return {
            self._build_var_name(
                "particle crack length distribution", "[m]"
            ): l_cr_dist,
            self._build_var_name(
                "particle crack length distribution", "[m]", averaged=True
            ): l_cr_av_dist,
        }

    def _get_mechanical_size_distribution_results(
        self, variables: dict[str, pybamm.Symbol]
    ) -> dict[str, pybamm.Symbol]:
        """
        Calculate mechanical results for particle size distributions.

        Computes stress and displacement for particle size distributions,
        following the models in Ai2019 and Deshpande2012.
        """
        domain, Domain = self.domain_Domain
        phase_name = self.phase_param.phase_name
        phase_param = self.phase_param

        # Extract concentration variables
        c_s_rav = variables[
            f"R-averaged {domain} {phase_name}"
            "particle concentration distribution [mol.m-3]"
        ]
        c_s_surf = variables[
            f"{Domain} {phase_name}"
            "particle surface concentration distribution [mol.m-3]"
        ]
        sto = variables[f"{Domain} {phase_name}particle concentration distribution"]

        # Broadcast temperature to particle size and particle domains
        T = self._broadcast_temperature_to_particle_size(variables)

        # Compute material properties (tangential approximation for Omega)
        Omega = pybamm.r_average(phase_param.Omega(sto, T))
        E = pybamm.r_average(phase_param.E(sto, T))

        # Calculate stress and displacement (Ai2019 equations)
        stress_disp = self._compute_distribution_stress_displacement(
            c_s_rav, c_s_surf, Omega, E, phase_param
        )

        # Add results to variables dictionary
        self._add_distribution_mechanics_variables(variables, stress_disp)

        return variables

    def _broadcast_temperature_to_particle_size(
        self, variables: dict[str, pybamm.Symbol]
    ) -> pybamm.Symbol:
        """Broadcast electrode temperature to particle size and particle domains."""
        domain, Domain = self.domain_Domain
        phase_name = self.phase_param.phase_name

        T = pybamm.PrimaryBroadcast(
            variables[f"{Domain} electrode temperature [K]"],
            [f"{domain} {phase_name}particle size"],
        )
        return pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle"])

    def _compute_distribution_stress_displacement(
        self,
        c_s_rav: pybamm.Symbol,
        c_s_surf: pybamm.Symbol,
        Omega: pybamm.Symbol,
        E: pybamm.Symbol,
        phase_param: Any,
    ) -> dict[str, pybamm.Symbol]:
        """
        Compute stress and displacement for size distributions.

        Based on Ai2019 equations:
        - Eq [7]: Radial stress at surface (r=R) is zero
        - Eq [8]: Tangential stress from concentration gradient
        - Eq [10]: Surface displacement from volume change
        """
        c_0 = phase_param.c_0
        R0 = phase_param.R
        nu = phase_param.nu

        # Surface displacement (Ai2019 eq [10])
        # c_0 is reference concentration for no deformation
        disp_surf = Omega * R0 * (c_s_rav - c_0) / self.DISPLACEMENT_GEOMETRIC_FACTOR

        # Surface stresses (Ai2019 eqs [7-8])
        # Radial stress at particle surface is zero (eq [7] with r=R)
        stress_r_surf = pybamm.Scalar(0)

        # Tangential stress (eq [8] with r=R)
        # Note: c_s_rav is already multiplied by 3/R^3 inside r_average
        stress_t_surf = (
            Omega * E * (c_s_rav - c_s_surf) / self.STRESS_GEOMETRIC_FACTOR / (1.0 - nu)
        )

        return {
            "stress_r_surf": stress_r_surf,
            "stress_t_surf": stress_t_surf,
            "disp_surf": disp_surf,
            "stress_r_surf_av": pybamm.x_average(stress_r_surf),
            "stress_t_surf_av": pybamm.x_average(stress_t_surf),
            "disp_surf_av": pybamm.x_average(disp_surf),
        }

    def _add_distribution_mechanics_variables(
        self, variables: dict[str, pybamm.Symbol], stress_disp: dict[str, pybamm.Symbol]
    ) -> None:
        """Add distribution mechanics variables to the variables dictionary."""
        domain, Domain = self.domain_Domain
        phase_name = self.phase_param.phase_name

        variables.update(
            {
                f"{Domain} {phase_name}"
                "particle surface radial stress distribution [Pa]": stress_disp[
                    "stress_r_surf"
                ],
                f"{Domain} {phase_name}"
                "particle surface tangential stress distribution [Pa]": stress_disp[
                    "stress_t_surf"
                ],
                f"{Domain} {phase_name}"
                "particle surface displacement distribution [m]": stress_disp[
                    "disp_surf"
                ],
                f"X-averaged {domain} {phase_name}"
                "particle surface radial stress distribution [Pa]": stress_disp[
                    "stress_r_surf_av"
                ],
                f"X-averaged {domain} {phase_name}"
                "particle surface tangential stress distribution [Pa]": stress_disp[
                    "stress_t_surf_av"
                ],
                f"X-averaged {domain} {phase_name}"
                "particle surface displacement distribution [m]": stress_disp[
                    "disp_surf_av"
                ],
            }
        )

    def _get_mechanical_results(
        self, variables: dict[str, pybamm.Symbol]
    ) -> dict[str, pybamm.Symbol]:
        """
        Calculate mechanical results including stress, displacement, and thickness changes.

        This method computes particle surface stress and displacement based on
        concentration gradients, following the mechanical models in Ai2019 and
        Deshpande2012.
        """
        # Extract required variables and compute particle mechanics
        mech_vars = self._extract_mechanical_variables(variables)
        stress_disp = self._compute_stress_and_displacement(mech_vars)

        # Update variables with computed results
        self._add_particle_mechanics_variables(variables, stress_disp)

        # Aggregate thickness changes at electrode and cell level
        self._aggregate_thickness_changes(variables, mech_vars)

        return variables

    def _extract_mechanical_variables(
        self, variables: dict[str, pybamm.Symbol]
    ) -> dict[str, Any]:
        """Extract and compute intermediate variables needed for mechanics calculations."""
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name
        phase_param = self.phase_param

        c_s_rav = variables[
            f"R-averaged {domain} {phase_name}particle concentration [mol.m-3]"
        ]
        sto_rav = variables[f"R-averaged {domain} {phase_name}particle concentration"]
        c_s_surf = variables[
            f"{Domain} {phase_name}particle surface concentration [mol.m-3]"
        ]
        T_xav = variables["X-averaged cell temperature [K]"]

        # Broadcast temperature to particle domain
        T = pybamm.PrimaryBroadcast(
            variables[f"{Domain} electrode temperature [K]"],
            [f"{domain} {phase_name}particle"],
        )
        eps_s = variables[
            f"{Domain} electrode {phase_name}active material volume fraction"
        ]
        sto = variables[f"{Domain} {phase_name}particle concentration"]

        # Compute material properties (tangential approximation for Omega)
        Omega = pybamm.r_average(phase_param.Omega(sto, T))
        E = pybamm.r_average(phase_param.E(sto, T))
        sto_init = pybamm.r_average(phase_param.c_init / phase_param.c_max)

        # Compute volume change for thickness calculation
        v_change = pybamm.x_average(
            eps_s * phase_param.t_change(sto_rav)
        ) - pybamm.x_average(eps_s * phase_param.t_change(sto_init))

        electrode_thickness_change = (
            self.param.n_electrodes_parallel * v_change * self.domain_param.L
        )

        return {
            "c_s_rav": c_s_rav,
            "c_s_surf": c_s_surf,
            "c_0": phase_param.c_0,
            "R": phase_param.R,
            "Omega": Omega,
            "E": E,
            "nu": phase_param.nu,
            "T_xav": T_xav,
            "electrode_thickness_change": electrode_thickness_change,
        }

    def _compute_stress_and_displacement(
        self, mech_vars: dict[str, Any]
    ) -> dict[str, pybamm.Symbol]:
        """
        Compute particle surface stress and displacement.

        Based on Ai2019 equations:
        - Eq [7]: Radial stress at surface (r=R) is zero
        - Eq [8]: Tangential stress from concentration gradient
        - Eq [10]: Surface displacement from volume change
        """
        # Surface displacement (Ai2019 eq [10])
        # Reference concentration c_0 gives zero deformation
        disp_surf = (
            mech_vars["Omega"]
            * mech_vars["R"]
            * (mech_vars["c_s_rav"] - mech_vars["c_0"])
            / self.DISPLACEMENT_GEOMETRIC_FACTOR
        )

        # Surface stresses (Ai2019 eqs [7-8])
        # Radial stress at particle surface is zero (eq [7] with r=R)
        stress_r_surf = pybamm.Scalar(0)

        # Tangential stress from concentration gradient (eq [8] with r=R)
        # Note: c_s_rav is already multiplied by 3/R^3 inside r_average
        stress_t_surf = (
            mech_vars["Omega"]
            * mech_vars["E"]
            * (mech_vars["c_s_rav"] - mech_vars["c_s_surf"])
            / self.STRESS_GEOMETRIC_FACTOR
            / (1.0 - mech_vars["nu"])
        )

        return {
            "stress_r_surf": stress_r_surf,
            "stress_t_surf": stress_t_surf,
            "disp_surf": disp_surf,
            "stress_r_surf_av": pybamm.x_average(stress_r_surf),
            "stress_t_surf_av": pybamm.x_average(stress_t_surf),
            "disp_surf_av": pybamm.x_average(disp_surf),
            "electrode_thickness_change": mech_vars["electrode_thickness_change"],
        }

    def _add_particle_mechanics_variables(
        self,
        variables: dict[str, pybamm.Symbol],
        stress_disp: dict[str, pybamm.Symbol],
    ) -> None:
        """Add computed stress and displacement variables to the variables dictionary."""
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        variables.update(
            {
                # Spatial distributions
                f"{Domain} {phase_name}particle surface radial stress [Pa]": stress_disp[
                    "stress_r_surf"
                ],
                f"{Domain} {phase_name}particle surface tangential stress [Pa]": stress_disp[
                    "stress_t_surf"
                ],
                f"{Domain} {phase_name}particle surface displacement [m]": stress_disp[
                    "disp_surf"
                ],
                # X-averaged quantities
                f"X-averaged {domain} {phase_name}particle surface radial stress [Pa]": stress_disp[
                    "stress_r_surf_av"
                ],
                f"X-averaged {domain} {phase_name}particle surface tangential stress [Pa]": stress_disp[
                    "stress_t_surf_av"
                ],
                f"X-averaged {domain} {phase_name}particle surface displacement [m]": stress_disp[
                    "disp_surf_av"
                ],
                # Thickness change
                f"{Domain} electrode {phase_name}thickness change [m]": stress_disp[
                    "electrode_thickness_change"
                ],
            }
        )

    def _aggregate_thickness_changes(
        self, variables: dict[str, pybamm.Symbol], mech_vars: dict[str, Any]
    ) -> None:
        """
        Aggregate thickness changes from phases to electrode level and from
        electrodes to cell level.
        """
        domain, Domain = self.domain_Domain

        # Aggregate primary and secondary phase thickness changes
        self._aggregate_phase_thickness_changes(variables, Domain)

        # Aggregate negative and positive electrode thickness changes to cell level
        self._aggregate_cell_thickness_change(variables, mech_vars["T_xav"])

    def _aggregate_phase_thickness_changes(
        self, variables: dict[str, pybamm.Symbol], Domain: str
    ) -> None:
        """Combine primary and secondary phase thickness changes for an electrode."""
        primary_key = f"{Domain} electrode primary thickness change [m]"
        secondary_key = f"{Domain} electrode secondary thickness change [m]"
        combined_key = f"{Domain} electrode thickness change [m]"

        if primary_key in variables and secondary_key in variables:
            variables[combined_key] = variables[primary_key] + variables[secondary_key]

    def _aggregate_cell_thickness_change(
        self, variables: dict[str, pybamm.Symbol], T_xav: pybamm.Symbol
    ) -> None:
        """
        Calculate total cell thickness change from electrode changes and thermal expansion.

        Based on Ai2019 eq [13] for thermal expansion contribution.
        """
        neg_key = "Negative electrode thickness change [m]"
        pos_key = "Positive electrode thickness change [m]"

        if neg_key in variables and pos_key in variables:
            # Thermal expansion contribution (Ai2019 eq [13])
            thermal_expansion = self.param.alpha_T_cell * (T_xav - self.param.T_ref)

            # Total cell thickness change
            variables["Cell thickness change [m]"] = (
                variables[neg_key] + variables[pos_key] + thermal_expansion
            )

    def _get_standard_surface_variables(
        self, variables: dict[str, pybamm.Symbol]
    ) -> dict[str, pybamm.Symbol]:
        """Calculate crack surface variables including roughness and surface area ratio."""
        domain, Domain = self.domain_Domain
        phase_name = self.phase_param.phase_name

        l_cr = variables[f"{Domain} {phase_name}particle crack length [m]"]
        a = variables[
            f"{Domain} electrode {phase_name}surface area to volume ratio [m-1]"
        ]
        rho_cr = self.phase_param.rho_cr
        w_cr = self.phase_param.w_cr

        # Roughness is ratio of cracks to normal surface
        roughness = 1 + self.CRACK_ROUGHNESS_FACTOR * l_cr * rho_cr * w_cr
        # Crack surface area to volume ratio
        a_cr = (roughness - 1) * a

        roughness_xavg = pybamm.x_average(roughness)
        return {
            f"{Domain} {phase_name}crack surface to volume ratio [m-1]": a_cr,
            f"{Domain} {phase_name}electrode roughness ratio": roughness,
            f"X-averaged {domain} {phase_name}electrode roughness ratio": roughness_xavg,
        }



#
# Base class for thermal effects
#
import numpy as np

import pybamm


class BaseThermal(pybamm.BaseSubModel):
    """
    Base class for thermal effects

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None, x_average=False):
        super().__init__(param, options=options)
        self.use_lumped_thermal_capacity = self.options.get(
            "use lumped thermal capacity", "false"
        )
        self.x_average = x_average

        if self.options["heat of mixing"] == "true":
            pybamm.citations.register("Richardson2021")

    def _get_standard_fundamental_variables(self, T_dict):
        """
        Note: here we explicitly pass in the averages for the temperature as computing
        the average temperature in `BaseThermal` using `self._x_average` requires a
        workaround to avoid raising a `ModelError` (as the key in the equation
        dict gets modified).

        For more information about this method in general,
        see :meth:`pybamm.base_submodel._get_standard_fundamental_variables`
        """

        # The variable T is the concatenation of the temperature in the middle domains
        # (e.g. negative electrode, separator and positive electrode for a full cell),
        # excluding current collectors, for use in the electrochemical models
        T_mid = [T_dict[k] for k in self.options.whole_cell_domains]
        T = pybamm.concatenation(*T_mid)

        # Get the ambient temperature, which can be specified as a function of space
        # (y, z) only and time
        y = pybamm.standard_spatial_vars.y
        z = pybamm.standard_spatial_vars.z
        T_amb = self.param.T_amb(y, z, pybamm.t)
        T_amb_av = self._yz_average(T_amb)

        variables = {
            "Ambient temperature [K]": T_amb,
            "Volume-averaged ambient temperature [K]": T_amb_av,
            "Cell temperature [K]": T,
        }
        for name, var in T_dict.items():
            Name = name.capitalize()
            variables[f"{Name} temperature [K]"] = var
            if name in ["negative electrode", "separator", "positive electrode"]:
                variables[f"X-averaged {name} temperature [K]"] = pybamm.x_average(var)

        # Calculate temperatures in Celsius
        variables_Kelvin = variables.copy()
        for name_K, var in variables_Kelvin.items():
            name_C = name_K.replace("[K]", "[C]")
            variables.update({name_C: var - 273.15})

        return variables

    def _get_standard_coupled_variables(self, variables):
        # Ohmic heating in solid
        i_s_p = variables["Positive electrode current density [A.m-2]"]
        phi_s_p = variables["Positive electrode potential [V]"]
        Q_ohm_s_cn, Q_ohm_s_cp = self._current_collector_heating(variables)
        if self.options.electrode_types["negative"] == "planar":
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            T_n = variables["Negative electrode temperature [K]"]
            Q_ohm_s_n = i_boundary_cc**2 / self.param.n.sigma(T_n)
        else:
            i_s_n = variables["Negative electrode current density [A.m-2]"]
            phi_s_n = variables["Negative electrode potential [V]"]
            Q_ohm_s_n = -pybamm.inner(i_s_n, pybamm.grad(phi_s_n))
        Q_ohm_s_s = pybamm.FullBroadcast(0, ["separator"], "current collector")
        Q_ohm_s_p = -pybamm.inner(i_s_p, pybamm.grad(phi_s_p))
        Q_ohm_s = pybamm.concatenation(Q_ohm_s_n, Q_ohm_s_s, Q_ohm_s_p)

        # Ohmic heating in electrolyte
        # TODO: change full stefan-maxwell conductivity so that i_e is always
        # a Concatenation
        i_e = variables["Electrolyte current density [A.m-2]"]
        # Special case for half cell -- i_e has to be a concatenation for this to work due to a mismatch with Q_ohm, so we make a new i_e which is a concatenation.
        if (not isinstance(i_e, pybamm.Concatenation)) and (
            self.options.electrode_types["negative"] == "planar"
        ):
            i_e = self._get_i_e_for_half_cell_thermal(variables)

        phi_e = variables["Electrolyte potential [V]"]
        if isinstance(i_e, pybamm.Concatenation):
            # compute by domain if possible
            phi_e_s = variables["Separator electrolyte potential [V]"]
            phi_e_p = variables["Positive electrolyte potential [V]"]
            if self.options.electrode_types["negative"] == "planar":
                i_e_s, i_e_p = i_e.orphans
                Q_ohm_e_n = pybamm.FullBroadcast(
                    0, ["negative electrode"], "current collector"
                )
            else:
                i_e_n, i_e_s, i_e_p = i_e.orphans
                phi_e_n = variables["Negative electrolyte potential [V]"]
                Q_ohm_e_n = -pybamm.inner(i_e_n, pybamm.grad(phi_e_n))
            Q_ohm_e_s = -pybamm.inner(i_e_s, pybamm.grad(phi_e_s))
            Q_ohm_e_p = -pybamm.inner(i_e_p, pybamm.grad(phi_e_p))
            Q_ohm_e = pybamm.concatenation(Q_ohm_e_n, Q_ohm_e_s, Q_ohm_e_p)
        else:
            # else compute using i_e across all domains
            if self.options.electrode_types["negative"] == "planar":
                Q_ohm_e_n = pybamm.FullBroadcast(
                    0, ["negative electrode"], "current collector"
                )
                Q_ohm_e_s_p = -pybamm.inner(i_e, pybamm.grad(phi_e))
                Q_ohm_e = pybamm.concatenation(Q_ohm_e_n, Q_ohm_e_s_p)
            else:
                Q_ohm_e = -pybamm.inner(i_e, pybamm.grad(phi_e))

        # Total Ohmic heating
        Q_ohm = Q_ohm_s + Q_ohm_e

        # Electrochemical heating (by electrode and phase)
        num_phases = int(self.options.positive["particle phases"])
        phase_names = [""]
        if num_phases > 1:
            phase_names = ["primary ", "secondary "]
        ocp_options = self.options.positive["open-circuit potential"]
        if isinstance(ocp_options, str):
            ocp_options = [ocp_options]

        Q_rxn_p, Q_rev_p, Q_hys_p = 0, 0, 0
        T_p = variables["Positive electrode temperature [K]"]
        for phase, ocp_option in zip(phase_names, ocp_options, strict=False):
            a_j_p = variables[
                f"Positive electrode {phase}volumetric interfacial current density [A.m-3]"
            ]
            eta_r_p = variables[f"Positive electrode {phase}reaction overpotential [V]"]
            # Irreversible electrochemical heating
            Q_rxn_p += a_j_p * eta_r_p
            # Reversible electrochemical heating
            dUdT_p = variables[f"Positive electrode {phase}entropic change [V.K-1]"]
            Q_rev_p += a_j_p * T_p * dUdT_p
            # Hysteresis electrochemical heating
            if ocp_option == "single":
                V_hys_p = 0
            else:
                pybamm.citations.register("Wycisk2023")
                U_p = variables[f"Positive electrode {phase}open-circuit potential [V]"]
                U_eq_p = variables[
                    f"Positive electrode {phase}equilibrium open-circuit potential [V]"
                ]
                V_hys_p = U_p - U_eq_p
            i_vol_p = variables[
                f"Positive electrode {phase}volumetric interfacial current density [A.m-3]"
            ]
            Q_hys_p += i_vol_p * V_hys_p

        num_phases = int(self.options.negative["particle phases"])
        phase_names = [""]
        if num_phases > 1:
            phase_names = ["primary ", "secondary "]
        ocp_options = self.options.negative["open-circuit potential"]
        if isinstance(ocp_options, str):
            ocp_options = [ocp_options]

        if self.options.electrode_types["negative"] == "planar":
            i_n = variables["Lithium metal total interfacial current density [A.m-2]"]
            eta_r_n = variables["Lithium metal interface reaction overpotential [V]"]
            Q_rxn_n = pybamm.PrimaryBroadcast(
                i_n * eta_r_n / self.param.n.L,
                ["negative electrode"],
                "current collector",
            )
            Q_rev_n = pybamm.FullBroadcast(
                0, ["negative electrode"], "current collector"
            )
            Q_hys_n = pybamm.FullBroadcast(
                0, ["negative electrode"], "current collector"
            )
        else:
            T_n = variables["Negative electrode temperature [K]"]
            Q_rxn_n, Q_rev_n, Q_hys_n = 0, 0, 0
            for phase, ocp_option in zip(phase_names, ocp_options, strict=False):
                a_j_n = variables[
                    f"Negative electrode {phase}volumetric interfacial current density [A.m-3]"
                ]
                eta_r_n = variables[
                    f"Negative electrode {phase}reaction overpotential [V]"
                ]
                # Irreversible electrochemical heating
                Q_rxn_n += a_j_n * eta_r_n
                # Reversible electrochemical heating
                dUdT_n = variables[f"Negative electrode {phase}entropic change [V.K-1]"]
                Q_rev_n += a_j_n * T_n * dUdT_n
                # Hysteresis electrochemical heating
                if ocp_option == "single":
                    V_hys_n = 0
                else:
                    pybamm.citations.register("Wycisk2023")
                    U_n = variables[
                        f"Negative electrode {phase}open-circuit potential [V]"
                    ]
                    U_eq_n = variables[
                        f"Negative electrode {phase}equilibrium open-circuit potential [V]"
                    ]
                    V_hys_n = U_n - U_eq_n
                i_vol_n = variables[
                    f"Negative electrode {phase}volumetric interfacial current density [A.m-3]"
                ]
                Q_hys_n += i_vol_n * V_hys_n

        # Irreversible electrochemical heating
        Q_rxn = pybamm.concatenation(
            Q_rxn_n, pybamm.FullBroadcast(0, "separator", "current collector"), Q_rxn_p
        )

        # Reversible electrochemical heating
        Q_rev = pybamm.concatenation(
            Q_rev_n, pybamm.FullBroadcast(0, "separator", "current collector"), Q_rev_p
        )

        # Heat of mixing
        Q_mix_s_n, Q_mix_s_s, Q_mix_s_p = self._heat_of_mixing(variables)
        Q_mix = pybamm.concatenation(Q_mix_s_n, Q_mix_s_s, Q_mix_s_p)

        # Heating due to hysteresis
        Q_hys = pybamm.concatenation(
            Q_hys_n, pybamm.FullBroadcast(0, "separator", "current collector"), Q_hys_p
        )

        # Total heating
        Q = Q_ohm + Q_rxn + Q_rev + Q_mix + Q_hys

        # Compute the X-average over the entire cell, including current collectors
        # Note: this can still be a function of y and z for higher-dimensional pouch
        # cell models
        Q_ohm_av = self._x_average(Q_ohm, Q_ohm_s_cn, Q_ohm_s_cp)
        Q_rxn_av = self._x_average(Q_rxn, 0, 0)
        Q_rev_av = self._x_average(Q_rev, 0, 0)
        Q_mix_av = self._x_average(Q_mix, 0, 0)
        Q_hys_av = self._x_average(Q_hys, 0, 0)
        Q_av = self._x_average(Q, Q_ohm_s_cn, Q_ohm_s_cp)

        # Compute the integrated heat source per unit simulated electrode-pair area
        # in W.m-2. Note: this can still be a function of y and z for
        # higher-dimensional pouch cell models
        Q_ohm_Wm2 = Q_ohm_av * self.param.L
        Q_rxn_Wm2 = Q_rxn_av * self.param.L
        Q_rev_Wm2 = Q_rev_av * self.param.L
        Q_mix_Wm2 = Q_mix_av * self.param.L
        Q_hys_Wm2 = Q_hys_av * self.param.L
        Q_Wm2 = Q_av * self.param.L

        # Now average over the electrode height and width
        Q_ohm_Wm2_av = self._yz_average(Q_ohm_Wm2)
        Q_rxn_Wm2_av = self._yz_average(Q_rxn_Wm2)
        Q_rev_Wm2_av = self._yz_average(Q_rev_Wm2)
        Q_mix_Wm2_av = self._yz_average(Q_mix_Wm2)
        Q_hys_Wm2_av = self._yz_average(Q_hys_Wm2)
        Q_Wm2_av = self._yz_average(Q_Wm2)

        # Compute total heat source terms (in W) over the *entire cell volume*, not
        # the product of electrode height * electrode width * electrode stack thickness
        # Note: we multiply by the number of electrode pairs, since the Q_xx_Wm2_av
        # variables are per electrode pair
        n_elec = self.param.n_electrodes_parallel
        A = self.param.L_y * self.param.L_z  # *modelled* electrode area
        Q_ohm_W = Q_ohm_Wm2_av * n_elec * A
        Q_rxn_W = Q_rxn_Wm2_av * n_elec * A
        Q_rev_W = Q_rev_Wm2_av * n_elec * A
        Q_mix_W = Q_mix_Wm2_av * n_elec * A
        Q_hys_W = Q_hys_Wm2_av * n_elec * A
        Q_W = Q_Wm2_av * n_elec * A

        # Compute volume-averaged heat source terms over the *entire cell volume*, not
        # the product of electrode height * electrode width * electrode stack thickness
        V = self.param.V_cell  # *actual* cell volume
        Q_ohm_vol_av = Q_ohm_W / V
        Q_rxn_vol_av = Q_rxn_W / V
        Q_rev_vol_av = Q_rev_W / V
        Q_mix_vol_av = Q_mix_W / V
        Q_hys_vol_av = Q_hys_W / V
        Q_vol_av = Q_W / V

        # Add lumped heat capacity if option is enabled
        if self.use_lumped_thermal_capacity == "true":
            rho_c_p_eff_av = self.param.cell_heat_capacity
        else:
            # Effective heat capacity
            T_vol_av = variables["Volume-averaged cell temperature [K]"]
            rho_c_p_eff_av = self.param.rho_c_p_eff(T_vol_av)

        variables.update(
            {
                # Ohmic
                "Ohmic heating [W.m-3]": Q_ohm,
                "X-averaged Ohmic heating [W.m-3]": Q_ohm_av,
                "Volume-averaged Ohmic heating [W.m-3]": Q_ohm_vol_av,
                "Volume-averaged heat of mixing [W.m-3]": Q_mix_vol_av,
                "Ohmic heating per unit electrode-pair area [W.m-2]": Q_ohm_Wm2,
                "Ohmic heating [W]": Q_ohm_W,
                # Irreversible
                "Irreversible electrochemical heating [W.m-3]": Q_rxn,
                "X-averaged irreversible electrochemical heating [W.m-3]": Q_rxn_av,
                "Volume-averaged irreversible electrochemical heating "
                + "[W.m-3]": Q_rxn_vol_av,
                "Irreversible electrochemical heating per unit "
                + "electrode-pair area [W.m-2]": Q_rxn_Wm2,
                "Irreversible electrochemical heating [W]": Q_rxn_W,
                # Reversible
                "Reversible heating [W.m-3]": Q_rev,
                "X-averaged reversible heating [W.m-3]": Q_rev_av,
                "Volume-averaged reversible heating [W.m-3]": Q_rev_vol_av,
                "Reversible heating per unit electrode-pair area [W.m-2]": Q_rev_Wm2,
                "Reversible heating [W]": Q_rev_W,
                # Mixing
                "Heat of mixing [W.m-3]": Q_mix,
                "X-averaged heat of mixing [W.m-3]": Q_mix_av,
                "Volume-averaged heating of mixing [W.m-3]": Q_mix_vol_av,
                "Heat of mixing per unit electrode-pair area [W.m-2]": Q_mix_Wm2,
                "Heat of mixing [W]": Q_mix_W,
                # Hysteresis
                "Hysteresis electrochemical heating [W.m-3]": Q_hys,
                "X-averaged hysteresis electrochemical heating [W.m-3]": Q_hys_av,
                "Volume-averaged hysteresis electrochemical heating [W.m-3]": Q_hys_vol_av,
                "Hysteresis electrochemical heating per unit electrode-pair area [W.m-2]": Q_hys_Wm2,
                "Hysteresis electrochemical heating [W]": Q_hys_W,
                # Total
                "Total heating [W.m-3]": Q,
                "X-averaged total heating [W.m-3]": Q_av,
                "Volume-averaged total heating [W.m-3]": Q_vol_av,
                "Total heating per unit electrode-pair area [W.m-2]": Q_Wm2,
                "Total heating [W]": Q_W,
                # Current collector
                "Negative current collector Ohmic heating [W.m-3]": Q_ohm_s_cn,
                "Positive current collector Ohmic heating [W.m-3]": Q_ohm_s_cp,
                # Effective heat capacity
                "Volume-averaged effective heat capacity [J.K-1.m-3]": rho_c_p_eff_av,
                "Cell thermal volume [m3]": V,
            }
        )
        return variables

    def _current_collector_heating(self, variables):
        """Compute Ohmic heating in current collectors."""
        cc_dimension = self.options["dimensionality"]

        # Compute the Ohmic heating for 0D current collectors
        if cc_dimension == 0:
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            Q_s_cn = i_boundary_cc**2 / self.param.n.sigma_cc
            Q_s_cp = i_boundary_cc**2 / self.param.p.sigma_cc
        # Otherwise we compute the Ohmic heating for 1 or 2D current collectors
        # In this limit the current flow is all in the y,z direction in the current
        # collectors
        elif cc_dimension in [1, 2]:
            phi_s_cn = variables["Negative current collector potential [V]"]
            phi_s_cp = variables["Positive current collector potential [V]"]
            # TODO: implement grad_squared in other spatial methods so that the
            # if statement can be removed
            if cc_dimension == 1:
                Q_s_cn = self.param.n.sigma_cc * pybamm.inner(
                    pybamm.grad(phi_s_cn), pybamm.grad(phi_s_cn)
                )
                Q_s_cp = self.param.p.sigma_cc * pybamm.inner(
                    pybamm.grad(phi_s_cp), pybamm.grad(phi_s_cp)
                )
            elif cc_dimension == 2:
                # Inner not implemented in 2D -- have to call grad_squared directly
                Q_s_cn = self.param.n.sigma_cc * pybamm.grad_squared(phi_s_cn)
                Q_s_cp = self.param.p.sigma_cc * pybamm.grad_squared(phi_s_cp)
        return Q_s_cn, Q_s_cp

    def _heat_of_mixing(self, variables):
        """Compute heat of mixing source terms."""

        if self.options["heat of mixing"] == "true":
            F = pybamm.constants.F.value
            pi = np.pi

            # Compute heat of mixing in negative electrode
            if self.options.electrode_types["negative"] == "planar":
                Q_mix_s_n = pybamm.FullBroadcast(
                    0, ["negative electrode"], "current collector"
                )
            else:
                a_n = variables["Negative electrode surface area to volume ratio [m-1]"]
                R_n = variables["Negative particle radius [m]"]
                N_n = a_n / (4 * pi * R_n**2)
                if self.x_average:
                    c_n = variables[
                        "X-averaged negative particle concentration [mol.m-3]"
                    ]
                    T_n = variables["X-averaged negative electrode temperature [K]"]
                else:
                    c_n = variables["Negative particle concentration [mol.m-3]"]
                    T_n = variables["Negative electrode temperature [K]"]
                T_n_part = pybamm.PrimaryBroadcast(T_n, ["negative particle"])
                dc_n_dr2 = pybamm.inner(pybamm.grad(c_n), pybamm.grad(c_n))
                D_n = self.param.n.prim.D(c_n, T_n_part)
                dUeq_n = self.param.n.prim.U(
                    c_n / self.param.n.prim.c_max, T_n_part
                ).diff(c_n)
                integrand_r_n = D_n * dc_n_dr2 * dUeq_n
                integration_variable_r_n = [
                    pybamm.SpatialVariable("r", domain=integrand_r_n.domain)
                ]
                integral_r_n = pybamm.Integral(integrand_r_n, integration_variable_r_n)
                Q_mix_s_n = -F * N_n * integral_r_n

            # Compute heat of mixing in positive electrode
            a_p = variables["Positive electrode surface area to volume ratio [m-1]"]
            R_p = variables["Positive particle radius [m]"]
            N_p = a_p / (4 * pi * R_p**2)
            if self.x_average:
                c_p = variables["X-averaged positive particle concentration [mol.m-3]"]
                T_p = variables["X-averaged positive electrode temperature [K]"]
            else:
                c_p = variables["Positive particle concentration [mol.m-3]"]
                T_p = variables["Positive electrode temperature [K]"]
            T_p_part = pybamm.PrimaryBroadcast(T_p, ["positive particle"])
            dc_p_dr2 = pybamm.inner(pybamm.grad(c_p), pybamm.grad(c_p))
            D_p = self.param.p.prim.D(c_p, T_p_part)
            dUeq_p = self.param.p.prim.U(c_p / self.param.p.prim.c_max, T_p_part).diff(
                c_p
            )
            integrand_r_p = D_p * dc_p_dr2 * dUeq_p
            integration_variable_r_p = [
                pybamm.SpatialVariable("r", domain=integrand_r_p.domain)
            ]
            integral_r_p = pybamm.Integral(integrand_r_p, integration_variable_r_p)
            Q_mix_s_p = -F * N_p * integral_r_p
            Q_mix_s_s = pybamm.FullBroadcast(0, ["separator"], "current collector")
        else:
            Q_mix_s_n = pybamm.FullBroadcast(
                0, ["negative electrode"], "current collector"
            )
            Q_mix_s_p = pybamm.FullBroadcast(
                0, ["positive electrode"], "current collector"
            )
            Q_mix_s_s = pybamm.FullBroadcast(0, ["separator"], "current collector")

        return Q_mix_s_n, Q_mix_s_s, Q_mix_s_p

    def _x_average(self, var, var_cn, var_cp):
        """
        Computes the X-average over the whole cell (including current collectors)
        from the variable in the cell (negative electrode, separator,
        positive electrode), negative current collector, and positive current
        collector.
        Note: we do this as we cannot create a single variable which is
        the concatenation [var_cn, var, var_cp] since var_cn and var_cp share the
        same domain. (In the N+1D formulation the current collector variables are
        assumed independent of x, so we do not make the distinction between negative
        and positive current collectors in the geometry).
        """
        out = (
            self.param.n.L_cc * var_cn
            + self.param.L_x * pybamm.x_average(var)
            + self.param.p.L_cc * var_cp
        ) / self.param.L
        return out

    def _yz_average(self, var):
        """Computes the y-z average."""
        # TODO: change the behaviour of z_average and yz_average so the if statement
        # can be removed
        if self.options["dimensionality"] in [0, 1]:
            return pybamm.z_average(var)
        elif self.options["dimensionality"] == 2:
            return pybamm.yz_average(var)

    def _get_i_e_for_half_cell_thermal(self, variables):
        c_e_s = variables["Separator electrolyte concentration [mol.m-3]"]
        c_e_p = variables["Positive electrolyte concentration [mol.m-3]"]
        grad_phi_e_s = variables["Gradient of separator electrolyte potential [V.m-1]"]
        grad_phi_e_p = variables["Gradient of positive electrolyte potential [V.m-1]"]
        grad_c_e_s = pybamm.grad(c_e_s)
        grad_c_e_p = pybamm.grad(c_e_p)
        T_s = variables["Separator temperature [K]"]
        T_p = variables["Positive electrode temperature [K]"]
        tor_s = variables["Separator electrolyte transport efficiency"]
        tor_p = variables["Positive electrolyte transport efficiency"]
        i_e_s = (self.param.kappa_e(c_e_s, T_s) * tor_s) * (
            self.param.chiRT_over_Fc(c_e_s, T_s) * grad_c_e_s - grad_phi_e_s
        )
        i_e_p = (self.param.kappa_e(c_e_p, T_p) * tor_p) * (
            self.param.chiRT_over_Fc(c_e_p, T_p) * grad_c_e_p - grad_phi_e_p
        )
        i_e = pybamm.concatenation(i_e_s, i_e_p)
        return i_e



#
# Many-Particle Model (MPM)
#
import pybamm

from .spm import SPM


class MPM(SPM):
    """
    Many-Particle Model (MPM) of a lithium-ion battery with particle-size
    distributions for each electrode, from :footcite:t:`Kirk2020`.
    See :class:`pybamm.lithium_ion.BaseModel` for more details.

    Examples
    --------
    >>> model = pybamm.lithium_ion.MPM()
    >>> model.name
    'Many-Particle Model'

    """

    def __init__(self, options=None, name="Many-Particle Model", build=True):
        # Necessary/default options
        options = options or {}
        if "particle size" in options and options["particle size"] != "distribution":
            raise pybamm.OptionError(
                "particle size must be 'distribution' for MPM not '{}'".format(
                    options["particle size"]
                )
            )
        elif "surface form" in options and options["surface form"] == "false":
            raise pybamm.OptionError(
                "surface form must be 'algebraic' or 'differential' for MPM not 'false'"
            )
        else:
            surface_form = options.get("surface form", "algebraic")
            options.update(
                {"particle size": "distribution", "surface form": surface_form}
            )
        super().__init__(options, name, build)

        pybamm.citations.register("Kirk2020")
        pybamm.citations.register("Kirk2021")

    @property
    def default_parameter_values(self):
        default_params = super().default_parameter_values
        default_params = pybamm.get_size_distribution_parameters(
            default_params, working_electrode=self.options["working electrode"]
        )
        return default_params



#
# Class for reaction driven porosity changes as a multiple of SEI/plating thicknesses
#
import pybamm

from .base_porosity import BaseModel


class ReactionDriven(BaseModel):
    """Reaction-driven porosity changes as a multiple of SEI/plating thicknesses

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        Options dictionary passed from the full model
    x_average : bool
        Whether to use x-averaged variables (SPM, SPMe, etc) or full variables (DFN)
    """

    def __init__(self, param, options, x_average):
        super().__init__(param, options)
        self.x_average = x_average

    def get_coupled_variables(self, variables):
        eps_dict = {}
        for domain in self.options.whole_cell_domains:
            delta_eps_k = 0
            if domain != "separator":  # separator porosity does not change
                dom = domain.split()[0]
                Domain = dom.capitalize()
                SEI_option = getattr(self.options, dom)["SEI"]
                phases_option = getattr(self.options, dom)["particle phases"]
                phases = self.options.phases[dom]
                for phase in phases:
                    if phases_option == "1" and phase == "primary":
                        # `domain` has one phase
                        phase_name = ""
                        pref = ""
                    else:
                        # `domain` has more than one phase
                        phase_name = phase + " "
                        pref = phase.capitalize() + ": "
                    a_k = variables[
                        f"{Domain} electrode {phase_name}"
                        "surface area to volume ratio [m-1]"
                    ]
                    if SEI_option == "none":
                        L_sei_0 = pybamm.Scalar(0)
                    else:
                        L_sei_0 = pybamm.Parameter(f"{pref}Initial SEI thickness [m]")
                    L_sei_k = variables[f"{Domain} {phase_name}SEI thickness [m]"]
                    L_pl_k = variables[
                        f"{Domain} {phase_name}lithium plating thickness [m]"
                    ]
                    L_dead_k = variables[
                        f"{Domain} {phase_name}dead lithium thickness [m]"
                    ]
                    L_sei_cr_k = variables[
                        f"{Domain} {phase_name}SEI on cracks thickness [m]"
                    ]
                    roughness_k = variables[
                        f"{Domain} {phase_name}electrode roughness ratio"
                    ]

                    L_tot = (
                        (L_sei_k - L_sei_0)
                        + L_pl_k
                        + L_dead_k
                        + L_sei_cr_k * (roughness_k - 1)
                    )

                    # This assumes a thin film so curvature effects are neglected.
                    # They could be included (e.g. for a sphere it is
                    # a_n * (L_tot + L_tot ** 2 / R_n + L_tot ** # 3 / (3 * R_n ** 2)))
                    # but it is not clear if it is relevant or not.
                    delta_eps_k += -a_k * L_tot

            domain_param = self.param.domain_params[domain.split()[0]]
            eps_k = domain_param.epsilon_init + delta_eps_k
            eps_dict[domain] = eps_k

        variables = self._get_standard_porosity_variables(eps_dict)

        return variables

    def add_events_from(self, variables):
        eps_p = variables["Positive electrode porosity"]
        self.events.append(
            pybamm.Event(
                "Zero positive electrode porosity cut-off",
                pybamm.min(eps_p),
                pybamm.EventType.TERMINATION,
            )
        )
        self.events.append(
            pybamm.Event(
                "Max positive electrode porosity cut-off",
                1 - pybamm.max(eps_p),
                pybamm.EventType.TERMINATION,
            )
        )
        if "negative electrode" in self.options.whole_cell_domains:
            eps_n = variables["Negative electrode porosity"]
            self.events.append(
                pybamm.Event(
                    "Zero negative electrode porosity cut-off",
                    pybamm.min(eps_n),
                    pybamm.EventType.TERMINATION,
                )
            )
            self.events.append(
                pybamm.Event(
                    "Max negative electrode porosity cut-off",
                    1 - pybamm.max(eps_n),
                    pybamm.EventType.TERMINATION,
                )
            )



#
# IndependentVariable class
#
from __future__ import annotations

import numpy as np
import numpy.typing as npt
import sympy

import pybamm
from pybamm.type_definitions import AuxiliaryDomainType, DomainsType, DomainType

KNOWN_COORD_SYS = ["cartesian", "cylindrical polar", "spherical polar"]


class IndependentVariable(pybamm.Symbol):
    """
    A node in the expression tree representing an independent variable.

    Used for expressing functions depending on a spatial variable or time

    Parameters
    ----------
    name : str
        name of the node
    domain : iterable of str
        list of domains that this variable is valid over
    auxiliary_domains : dict, optional
        dictionary of auxiliary domains, defaults to empty dict
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    """

    def __init__(
        self,
        name: str,
        domain: DomainType = None,
        auxiliary_domains: AuxiliaryDomainType = None,
        domains: DomainsType = None,
    ) -> None:
        super().__init__(
            name, domain=domain, auxiliary_domains=auxiliary_domains, domains=domains
        )

    @classmethod
    def _from_json(cls, snippet: dict):
        return cls(snippet["name"], domains=snippet["domains"])

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(self.domains)

    def _jac(self, variable) -> pybamm.Scalar:
        """See :meth:`pybamm.Symbol._jac()`."""
        return pybamm.Scalar(0)

    def to_equation(self) -> sympy.Symbol:
        """Convert the node and its subtree into a SymPy equation."""
        if self.print_name is not None:
            return sympy.Symbol(self.print_name)
        else:
            return sympy.Symbol(self.name)

    def create_copy(
        self,
        new_children=None,
        perform_simplifications=True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return self.__class__(self.name, domains=self.domains)


class Time(IndependentVariable):
    """
    A node in the expression tree representing time.
    """

    def __init__(self):
        super().__init__("time")

    @classmethod
    def _from_json(cls, snippet: dict):
        return cls()

    def create_copy(
        self,
        new_children=None,
        perform_simplifications=True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return Time()

    def _base_evaluate(
        self,
        t: float | None = None,
        y: npt.NDArray[np.float64] | None = None,
        y_dot: npt.NDArray[np.float64] | None = None,
        inputs: dict | str | None = None,
    ):
        """See :meth:`pybamm.Symbol._base_evaluate()`."""
        if t is None:
            raise ValueError("t must be provided")
        return t

    def _evaluate_for_shape(self):
        """
        Return the scalar '0' to represent the shape of the independent variable `Time`.
        See :meth:`pybamm.Symbol.evaluate_for_shape()`
        """
        return 0

    def to_equation(self):
        """Convert the node and its subtree into a SymPy equation."""
        return sympy.Symbol("t")


class SpatialVariable(IndependentVariable):
    """
    A node in the expression tree representing a spatial variable.

    Parameters
    ----------
    name : str
        name of the node (e.g. "x", "y", "z", "r", "x_n", "x_s", "x_p", "r_n", "r_p")
    domain : iterable of str
        list of domains that this variable is valid over (e.g. "cartesian", "spherical
        polar")
    auxiliary_domains : dict, optional
        dictionary of auxiliary domains, defaults to empty dict
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    """

    def __init__(
        self,
        name: str,
        domain: DomainType = None,
        auxiliary_domains: AuxiliaryDomainType = None,
        domains: DomainsType = None,
        coord_sys=None,
        direction=None,
    ) -> None:
        self.coord_sys = coord_sys
        super().__init__(
            name, domain=domain, auxiliary_domains=auxiliary_domains, domains=domains
        )
        domain = self.domain
        self.direction = direction

        if domain == []:
            raise ValueError("domain must be provided")

        # Check symbol name vs domain name
        if name == "r_n" and not all(n in domain[0] for n in ["negative", "particle"]):
            # catches "negative particle", "negative secondary particle", etc
            raise pybamm.DomainError(
                "domain must be negative particle if name is 'r_n'"
            )
        elif name == "r_p" and not all(
            n in domain[0] for n in ["positive", "particle"]
        ):
            # catches "positive particle", "positive secondary particle", etc
            raise pybamm.DomainError(
                "domain must be positive particle if name is 'r_p'"
            )
        elif name in ["x", "y", "z", "x_n", "x_s", "x_p"] and any(
            ["particle" in dom for dom in domain]
        ):
            raise pybamm.DomainError(f"domain cannot be particle if name is '{name}'")

    def create_copy(
        self,
        new_children=None,
        perform_simplifications=True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return self.__class__(self.name, domains=self.domains, coord_sys=self.coord_sys)


class SpatialVariableEdge(SpatialVariable):
    """
    A node in the expression tree representing a spatial variable, which evaluates
    on the edges

    Parameters
    ----------
    name : str
        name of the node (e.g. "x", "y", "z", "r", "x_n", "x_s", "x_p", "r_n", "r_p")
    domain : iterable of str
        list of domains that this variable is valid over (e.g. "cartesian", "spherical
        polar")
    auxiliary_domains : dict, optional
        dictionary of auxiliary domains, defaults to empty dict
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    """

    def __init__(
        self,
        name: str,
        domain: DomainType = None,
        auxiliary_domains: AuxiliaryDomainType = None,
        domains: DomainsType = None,
        coord_sys=None,
    ) -> None:
        super().__init__(name, domain, auxiliary_domains, domains, coord_sys)

    def _evaluates_on_edges(self, dimension):
        return True


# the independent variable time
t = Time()



#
# Variable class
#
from __future__ import annotations

import numbers

import numpy as np
import sympy

import pybamm
from pybamm.type_definitions import (
    AuxiliaryDomainType,
    DomainsType,
    DomainType,
    Numeric,
)


class VariableBase(pybamm.Symbol):
    """
    A node in the expression tree represending a dependent variable.

    This node will be discretised by :class:`.Discretisation` and converted
    to a :class:`pybamm.StateVector` node.

    Parameters
    ----------
    name : str
        name of the node
    domain : iterable of str
        list of domains that this variable is valid over
    auxiliary_domains : dict
        dictionary of auxiliary domains ({'secondary': ..., 'tertiary': ...,
        'quaternary': ...}). For example, for the single particle model, the particle
        concentration would be a Variable with domain 'negative particle' and secondary
        auxiliary domain 'current collector'. For the DFN, the particle concentration
        would be a Variable with domain 'negative particle', secondary domain
        'negative electrode' and tertiary domain 'current collector'
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    bounds : tuple, optional
        Physical bounds on the variable
    print_name : str, optional
        The name to use for printing. Default is None, in which case self.name is used.
    scale : float or :class:`pybamm.Symbol`, optional
        The scale of the variable, used for scaling the model when solving. The state
        vector representing this variable will be multiplied by this scale.
        Default is 1.
    reference : float or :class:`pybamm.Symbol`, optional
        The reference value of the variable, used for scaling the model when solving.
        This value will be added to the state vector representing this variable.
        Default is 0.
    """

    def __init__(
        self,
        name: str,
        domain: DomainType = None,
        auxiliary_domains: AuxiliaryDomainType = None,
        domains: DomainsType = None,
        bounds: tuple[pybamm.Symbol] | None = None,
        print_name: str | None = None,
        scale: float | pybamm.Symbol | None = 1,
        reference: float | pybamm.Symbol | None = 0,
    ):
        if isinstance(scale, numbers.Number):
            scale = pybamm.Scalar(scale)
        if isinstance(reference, numbers.Number):
            reference = pybamm.Scalar(reference)
        self._scale = scale
        self._reference = reference
        super().__init__(
            name,
            domain=domain,
            auxiliary_domains=auxiliary_domains,
            domains=domains,
        )
        self.bounds = bounds

        if print_name is None:
            print_name = name  # use name by default
        self.print_name = print_name

    @property
    def bounds(self):
        """Physical bounds on the variable."""
        return self._bounds

    @bounds.setter
    def bounds(self, values: tuple[Numeric, Numeric]):
        if values is None:
            values = (-np.inf, np.inf)
        else:
            if (
                all(isinstance(b, numbers.Number) for b in values)
                and values[0] >= values[1]
            ):
                raise ValueError(
                    f"Invalid bounds {values}. "
                    + "Lower bound should be strictly less than upper bound."
                )

        values = list(values)
        for idx, bound in enumerate(values):
            if isinstance(bound, numbers.Number):
                values[idx] = pybamm.Scalar(bound)
        self._bounds = tuple(values)

    def set_id(self):
        self._id = hash(
            (
                self.__class__,
                self.name,
                self.scale,
                self.reference,
                *tuple([(k, tuple(v)) for k, v in self.domains.items() if v != []]),
            )
        )

    def create_copy(
        self,
        new_children=None,
        perform_simplifications=True,
    ):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return self.__class__(
            self.name,
            domains=self.domains,
            bounds=self.bounds,
            print_name=self._raw_print_name,
            scale=self.scale,
            reference=self.reference,
        )

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(self.domains)

    def to_equation(self):
        """Convert the node and its subtree into a SymPy equation."""
        if self.print_name is not None:
            return sympy.Symbol(self.print_name)
        else:
            return self.name

    def to_json(
        self,
    ):
        raise NotImplementedError(
            "pybamm.Variable: Serialisation is only implemented for discretised models."
        )


class Variable(VariableBase):
    """
    A node in the expression tree represending a dependent variable.

    This node will be discretised by :class:`.Discretisation` and converted
    to a :class:`pybamm.StateVector` node.

    Parameters
    ----------

    name : str
        name of the node
        domain : iterable of str, optional
        list of domains that this variable is valid over
    auxiliary_domains : dict, optional
        dictionary of auxiliary domains ({'secondary': ..., 'tertiary': ...,
        'quaternary': ...}). For example, for the single particle model, the particle
        concentration would be a Variable with domain 'negative particle' and secondary
        auxiliary domain 'current collector'. For the DFN, the particle concentration
        would be a Variable with domain 'negative particle', secondary domain
        'negative electrode' and tertiary domain 'current collector'
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    bounds : tuple, optional
        Physical bounds on the variable
    print_name : str, optional
        The name to use for printing. Default is None, in which case self.name is used.
    scale : float or :class:`pybamm.Symbol`, optional
        The scale of the variable, used for scaling the model when solving. The state
        vector representing this variable will be multiplied by this scale.
        Default is 1.
    reference : float or :class:`pybamm.Symbol`, optional
        The reference value of the variable, used for scaling the model when solving.
        This value will be added to the state vector representing this variable.
        Default is 0.
    """

    def diff(self, variable: pybamm.Symbol):
        if variable == self:
            return pybamm.Scalar(1)
        elif variable == pybamm.t:
            # reference gets differentiated out
            return pybamm.VariableDot(
                self.name + "'", domains=self.domains, scale=self.scale
            )
        else:
            return pybamm.Scalar(0)


class VariableDot(VariableBase):
    """
    A node in the expression tree represending the time derviative of a dependent
    variable

    This node will be discretised by :class:`.Discretisation` and converted
    to a :class:`pybamm.StateVectorDot` node.

    Parameters
    ----------

    name : str
        name of the node
    domain : iterable of str
        list of domains that this variable is valid over
    auxiliary_domains : dict
        dictionary of auxiliary domains ({'secondary': ..., 'tertiary': ...,
        'quaternary': ...}). For example, for the single particle model, the particle
        concentration would be a Variable with domain 'negative particle' and secondary
        auxiliary domain 'current collector'. For the DFN, the particle concentration
        would be a Variable with domain 'negative particle', secondary domain
        'negative electrode' and tertiary domain 'current collector'
    domains : dict
        A dictionary equivalent to {'primary': domain, auxiliary_domains}. Either
        'domain' and 'auxiliary_domains', or just 'domains', should be provided
        (not both). In future, the 'domain' and 'auxiliary_domains' arguments may be
        deprecated.
    bounds : tuple, optional
        Physical bounds on the variable. Included for compatibility with `VariableBase`,
        but ignored.
    print_name : str, optional
        The name to use for printing. Default is None, in which case self.name is used.
    scale : float or :class:`pybamm.Symbol`, optional
        The scale of the variable, used for scaling the model when solving. The state
        vector representing this variable will be multiplied by this scale.
        Default is 1.
    reference : float or :class:`pybamm.Symbol`, optional
        The reference value of the variable, used for scaling the model when solving.
        This value will be added to the state vector representing this variable.
        Default is 0.
    """

    def get_variable(self) -> pybamm.Variable:
        """
        return a :class:`.Variable` corresponding to this VariableDot

        Note: Variable._jac adds a dash to the name of the corresponding VariableDot, so
        we remove this here
        """
        return Variable(self.name[:-1], domains=self.domains, scale=self.scale)

    def diff(self, variable: pybamm.Symbol) -> pybamm.Scalar:
        if variable == self:
            return pybamm.Scalar(1)
        elif variable == pybamm.t:
            raise pybamm.ModelError("cannot take second time derivative of a Variable")
        else:
            return pybamm.Scalar(0)
